@@ -3,7 +3,7 @@
 """
 Driver script for running the `forwardPremium` application on SMSCG.
 """
-# Copyright (C) 2011, 2012 University of Zurich. All rights reserved.
+# Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -19,7 +19,6 @@ Driver script for running the `forwardPremium` application on SMSCG.
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 #
-__version__ = '$Revision$'
 __author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
 # summary of user-visible changes
 __changelog__ = """
@@ -31,6 +30,7 @@ __changelog__ = """
 """
 __docformat__ = 'reStructuredText'
 
+from __future__ import absolute_import, print_function
 import shutil
 
 # Remove all files in curPath if -N option specified.
@@ -41,20 +41,19 @@ if __name__ == '__main__':
         path2Pymods = os.path.join(os.path.dirname(__file__), '../')
         if not sys.path.count(path2Pymods):
             sys.path.append(path2Pymods)
-        from pymods.support.support import rmFilesAndFolders
         curPath = os.getcwd()
         filesAndFolder = os.listdir(curPath)
         if 'gpremium.csv' in filesAndFolder: # if another paraSearch was run in here before, clean up.
             if 'para.loop' in os.listdir(os.getcwd()):
                 shutil.copyfile(os.path.join(curPath, 'para.loop'), os.path.join('/tmp', 'para.loop'))
-                rmFilesAndFolders(curPath)
+                shutil.rmtree(curPath)
                 shutil.copyfile(os.path.join('/tmp', 'para.loop'), os.path.join(curPath, 'para.loop'))
             else:
-                rmFilesAndFolders(curPath)
+                shutil.rmtree(curPath)
 
 
 # ugly workaround for Issue 95,
-# see: http://code.google.com/p/gc3pie/issues/detail?id=95
+# see: https://github.com/uzh/gc3pie/issues/95
 if __name__ == "__main__":
     import gpremium
 
@@ -128,7 +127,7 @@ Read `.loop` files and execute the `forwardPremium` program accordingly.
         if os.path.isdir(self.params.executable):
             self.params.executable = os.path.join(self.params.executable,
                                                   'forwardPremium')
-        gc3libs.utils.test_file(self.params.executable, os.R_OK|os.X_OK,
+        gc3libs.utils.check_file_access(self.params.executable, os.R_OK|os.X_OK,
                                 gc3libs.exceptions.InvalidUsage)
 
 
@@ -139,7 +138,7 @@ Read `.loop` files and execute the `forwardPremium` program accordingly.
         # Copy base dir
         localBaseDir = os.path.join(os.getcwd(), 'localBaseDir')
 #        gc3libs.utils.copytree(self.params.initial, '/mnt/shareOffice/ForwardPremium/Results/sensitivity/wGridSize/dfs')
-        gc3libs.utils.copytree(self.params.initial, localBaseDir)
+        shutil.copytree(self.params.initial, localBaseDir)
 
         for path in inputs:
             para_loop = path
