@@ -2,7 +2,7 @@
 #
 #   ggamess_retry.py -- Front-end script for submitting multiple GAMESS jobs to SMSCG.
 #
-#   Copyright (C) 2010, 2011, 2012 GC3, University of Zurich
+#   Copyright (C) 2010, 2011, 2012  University of Zurich. All rights reserved.
 #
 #   This program is free software: you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -23,7 +23,9 @@ It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.
 
 See the output of ``ggamess --help`` for program usage instructions.
 """
-__version__ = '2.2.3 version (SVN $Revision$)'
+
+from __future__ import absolute_import, print_function
+
 # summary of user-visible changes
 __changelog__ = """
   2011-08-03:
@@ -36,7 +38,7 @@ __docformat__ = 'reStructuredText'
 
 
 # workaround for Issue 95,
-# see: http://code.google.com/p/gc3pie/issues/detail?id=95
+# see: https://github.com/uzh/gc3pie/issues/95
 if __name__ == "__main__":
     import ggamess_retry
 
