@@ -2,7 +2,7 @@
 """
 Authentication support for the GC3Libs.
 """
-# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
+# Copyright (C) 2009-2015  University of Zurich. All rights reserved.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -15,17 +15,18 @@ Authentication support for the GC3Libs.
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+from __future__ import absolute_import, print_function, unicode_literals
+from builtins import str
+from builtins import object
 __docformat__ = 'reStructuredText'
-__version__ = '2.2.3 version (SVN $Revision: 4001 $)'
-
-import sys
 
 import gc3libs.exceptions
 
+
 class Auth(object):
+
     """
     A mish-mash of authorization functions.
 
@@ -62,15 +63,15 @@ class Auth(object):
         part, which is also hard-coded in the auth classes, and should not be.
     """
     types = {}
+
     def __init__(self, config, auto_enable):
         self.auto_enable = auto_enable
-        self.__auths = { }
+        self.__auths = {}
         self._config = config
-        self._ctors = { }
-        for auth_name, auth_params in self._config.iteritems():
+        self._ctors = {}
+        for auth_name, auth_params in self._config.items():
             self._ctors[auth_name] = Auth.types[auth_params['type']]
 
-
     def add_params(self, **params):
         """
         Add the specified keyword arguments as initialization
@@ -79,10 +80,9 @@ class Auth(object):
         Parameters that have already been specified are silently
         overwritten.
         """
-        for auth_name, auth_params in self._config.iteritems():
+        for auth_name, auth_params in self._config.items():
             auth_params.update(params)
 
-
     def get(self, auth_name, **kwargs):
         """
         Return an instance of the `Auth` class corresponding to the
@@ -108,8 +108,11 @@ class Auth(object):
             try:
                 params = self._config[auth_name].copy()
                 params.update(kwargs)
-                a =  self._ctors[auth_name](**dict(params))
-            except (AssertionError, AttributeError), ex:
+                a = self._ctors[auth_name](**dict(params))
+            except KeyError as err:
+                a = gc3libs.exceptions.ConfigurationError(
+                    "Unknown auth section %s" % (str(err),))
+            except (AssertionError, AttributeError) as ex:
                 a = gc3libs.exceptions.ConfigurationError(
                     "Missing required configuration parameters"
                     " in auth section '%s': %s" % (auth_name, str(ex)))
@@ -125,12 +128,13 @@ class Auth(object):
             if self.auto_enable:
                 try:
                     a.enable()
-                except gc3libs.exceptions.RecoverableAuthError, x:
+                except gc3libs.exceptions.RecoverableAuthError as x:
                     raise
-                except gc3libs.exceptions.UnrecoverableAuthError, x:
-                    gc3libs.log.debug("Got exception while enabling auth '%s',"
-                                      " will remember for next invocations:"
-                                      " %s: %s" % (auth_name, x.__class__.__name__, x))
+                except gc3libs.exceptions.UnrecoverableAuthError as x:
+                    gc3libs.log.debug(
+                        "Got exception while enabling auth '%s',"
+                        " will remember for next invocations:"
+                        " %s: %s" % (auth_name, x.__class__.__name__, x))
                     a = x
             else:
                 a = gc3libs.exceptions.UnrecoverableAuthError(
@@ -140,21 +144,22 @@ class Auth(object):
         self.__auths[auth_name] = a
         return a
 
-
     @staticmethod
     def register(auth_type, ctor):
         Auth.types[auth_type] = ctor
 
 
 class NoneAuth(object):
+
     """Auth proxy to use when no auth is needed."""
+
     def __init__(self, **auth):
         try:
             # test validity
             assert auth['type'] == 'none', (
                 "Configuration error. Unknown type: %s. Valid type: none"
                 % auth.type)
-        except AssertionError, x:
+        except AssertionError as x:
             raise gc3libs.exceptions.ConfigurationError(
                 'Erroneous configuration parameter: %s' % str(x))
 
@@ -162,7 +167,6 @@ class NoneAuth(object):
         return True
 
     def check(self):
-        gc3libs.log.debug("Checking auth: NoneAuth (always successful).")
         return True
 
     def enable(self):
@@ -171,12 +175,11 @@ class NoneAuth(object):
 Auth.register('none', NoneAuth)
 # register additional auth types
 # FIXME: it would be nice to have some kind of auto-discovery instead
-import gc3libs.authentication.grid
 import gc3libs.authentication.ssh
 import gc3libs.authentication.ec2
 import gc3libs.authentication.openstack
 
-## main: run tests
+# main: run tests
 
 if "__main__" == __name__:
     import doctest
