@@ -1,10 +1,10 @@
 #! /usr/bin/env python
-#
+
 """
 This module provides a generic BatchSystem class from which all
 batch-like backends should inherit.
 """
-# Copyright (C) 2009-2014 GC3, University of Zurich. All rights reserved.
+# Copyright (C) 2009-2019   University of Zurich. All rights reserved.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -17,27 +17,28 @@ batch-like backends should inherit.
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
-#
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+from __future__ import absolute_import, print_function, unicode_literals
+from builtins import str
 __docformat__ = 'reStructuredText'
-__version__ = '2.2.3 version (SVN $Revision$)'
 
 
+from collections import namedtuple
 from getpass import getuser
 import os
 import posixpath
-import random
 import shlex
-import stat
 import sys
 import tempfile
 import time
+import uuid
 
 import gc3libs
 from gc3libs import log, Run
+import gc3libs.defaults
 from gc3libs.backends import LRMS
-from gc3libs.utils import ifelse, same_docstring_as
+from gc3libs.utils import same_docstring_as, sh_quote_safe
 import gc3libs.backends.transport
 
 # Define some commonly used functions
@@ -95,6 +96,7 @@ def _make_remote_and_local_path_pair(transport, job, remote_relpath,
 
 
 class BatchSystem(LRMS):
+
     """
     Base class for backends dealing with a batch-queue system (e.g.,
     PBS/TORQUE, Grid Engine, etc.)
@@ -117,8 +119,16 @@ class BatchSystem(LRMS):
                  max_memory_per_core, max_walltime,
                  auth,  # ignored if `transport` is 'local'
                  # these are specific to the this backend
-                 frontend, transport, keyfile=None,
+                 frontend, transport,
                  accounting_delay=15,
+                 # SSH-related options; ignored if `transport` is 'local'
+                 ssh_config=None,
+                 keyfile=None,
+                 ignore_ssh_host_keys=False,
+                 ssh_timeout=None,
+                 large_file_threshold=None,
+                 large_file_chunk_size=None,
+                 spooldir=gc3libs.defaults.SPOOLDIR,
                  **extra_args):
 
         # init base class
@@ -129,17 +139,29 @@ class BatchSystem(LRMS):
 
         # backend-specific setup
         self.frontend = frontend
+        self.spooldir = spooldir
         if transport == 'local':
             self.transport = gc3libs.backends.transport.LocalTransport()
             self._username = getuser()
         elif transport == 'ssh':
             auth = self._auth_fn()
-            self._username = auth.username
+            self._username=auth.username
             self.transport = gc3libs.backends.transport.SshTransport(
-                frontend, username=self._username, keyfile=keyfile)
+                frontend,
+                ignore_ssh_host_keys=ignore_ssh_host_keys,
+                ssh_config=(ssh_config or auth.ssh_config),
+                username=self._username,
+                port=auth.port,
+                keyfile=(keyfile or auth.keyfile),
+                timeout=(ssh_timeout or auth.timeout),
+                large_file_threshold=large_file_threshold,
+                large_file_chunk_size=large_file_chunk_size,
+            )
         else:
             raise gc3libs.exceptions.TransportError(
                 "Unknown transport '%s'" % transport)
+        self.accounting_delay = accounting_delay
+
 
     def get_jobid_from_submit_output(self, output, regexp):
         """Parse the output of the submission command. Regexp is
@@ -149,8 +171,8 @@ class BatchSystem(LRMS):
             if match:
                 return match.group('jobid')
         raise gc3libs.exceptions.InternalError(
-            "Could not extract jobid from qsub output '%s'"
-            % output.rstrip())
+            "Could not extract jobid from submission command output `{0}`"
+            .format(output.rstrip()))
 
     def _get_command_argv(self, name, default=None):
         """
@@ -219,7 +241,7 @@ class BatchSystem(LRMS):
         # lookup the command name in the resource config parameters;
         # return it unchanged as a default
         argv = self._get_command_argv(name, default)
-        return str.join(' ', argv)
+        return ' '.join(argv)
 
     def _submit_command(self, app):
         """This method returns a string containing the command to
@@ -243,17 +265,36 @@ class BatchSystem(LRMS):
             "this should have been defined in a derived class.")
 
     def _parse_stat_output(self, stdout):
-        """This method will parse the output of the stat command and
-        return the current status of the job. The return value will be
-        a dictionary which will be used to update job's information.
+        """
+        Parse the output of the "stat" command and return the current
+        status of the job and its exit status.
 
-        The only expected key is `state`, which must be a valid
-        `Run.State` state.
+        :return: A `_stat_result`:class: instance.
         """
         raise NotImplementedError(
             "Abstract method `_parse_stat_output()` called - "
             "this should have been defined in a derived class.")
 
+    # the only purpose of this class definition is to add a docstring
+    # to the named tuple
+    class _stat_result(namedtuple('_stat_result', ['state', 'termstatus'])):
+        """
+        Result of parsing the "stat" command output.
+        It is a pair ``(state, termstatus)``, where:
+
+        * ``state`` is a valid GC3Pie task state
+          (i.e., instance of `Run.State`:class:, which see);
+
+        * ``termstatus`` is either ``None`` (to indicate that no
+          termination status information is available), or any value
+          that can be used to set GC3Pie's `Run.returncode`:attr: (for
+          instance, a 16-bit POSIX process termination status).
+
+        The ``termstatus`` part is only meaningful when ``state``
+        is ``TERMINATING``.
+        """
+        pass
+
     def _acct_command(self, job):
         """
         Return a string containing the command to issue to get accounting
@@ -318,8 +359,9 @@ class BatchSystem(LRMS):
                     "%s script not defined for resource %s", script, self.name)
                 continue
             if os.path.isfile(self[script]):
-                gc3libs.log.debug("Adding %s file `%s` to the submission script"
-                                  % (script, self[script]))
+                gc3libs.log.debug(
+                    "Adding %s file `%s` to the submission script"
+                    % (script, self[script]))
                 script_file = open(self[script])
                 script_txt.append("\n# %s file `%s` BEGIN\n"
                                   % (script, self[script]))
@@ -332,7 +374,7 @@ class BatchSystem(LRMS):
                 script_txt.append('\n# inline script BEGIN\n')
                 script_txt.append(self[script])
                 script_txt.append('\n# inline script END\n')
-        return str.join("", script_txt)
+        return "".join(script_txt)
 
     def get_prologue_script(self, app):
         """
@@ -340,8 +382,9 @@ class BatchSystem(LRMS):
         application and will return a string which contains the
         contents of the script(s) merged together.
         """
-        prologues = ['prologue', app.application_name+'_prologue',
-                     'prologue_content', app.application_name+'_prologue_content']
+        prologues = ['prologue', app.application_name + '_prologue',
+                     'prologue_content',
+                     app.application_name + '_prologue_content']
         return self._get_prepost_scripts(app, prologues)
 
     def get_epilogue_script(self, app):
@@ -350,8 +393,9 @@ class BatchSystem(LRMS):
         application and will return a string which contains the
         contents of the script(s) merged together.
         """
-        epilogues = ['epilogue', app.application_name+'epilogue',
-                     'epilogue_content', app.application_name+'epilogue_content']
+        epilogues = ['epilogue', app.application_name + '_epilogue',
+                     'epilogue_content',
+                     app.application_name + '_epilogue_content']
         return self._get_prepost_scripts(app, epilogues)
 
     @LRMS.authenticated
@@ -360,27 +404,24 @@ class BatchSystem(LRMS):
         sandbox, and will copy the sandbox in there.
         """
         job = app.execution
-        # Create the remote directory.
-        try:
-            self.transport.connect()
 
-            cmd = "mkdir -p $HOME/.gc3pie_jobs;" \
-                " mktemp -d $HOME/.gc3pie_jobs/lrms_job.XXXXXXXXXX"
-            log.info("Creating remote temporary folder: command '%s' " % cmd)
-            exit_code, stdout, stderr = self.transport.execute_command(cmd)
-            if exit_code == 0:
-                ssh_remote_folder = stdout.split('\n')[0]
-            else:
-                raise gc3libs.exceptions.LRMSError(
-                    "Failed executing command '%s' on resource '%s';"
-                    " exit code: %d, stderr: '%s'."
-                    % (cmd, self.name, exit_code, stderr))
-        except gc3libs.exceptions.TransportError, x:
-            raise
-        except:
-            raise
-                # Copy the input file to remote directory.
-        for local_path, remote_path in app.inputs.items():
+        # Create the remote directory.
+        self.transport.connect()
+        cmd = (
+            "mkdir -p {0};"
+            " mktemp -d {0}/batch_job.XXXXXXXXXX"
+            .format(self.spooldir))
+        exit_code, stdout, stderr = self.transport.execute_command(cmd)
+        if exit_code != 0:
+            raise gc3libs.exceptions.SpoolDirError(
+                "Cannot create temporary job working directory"
+                " on resource '%s'; command '%s' exited"
+                " with code: %d and stderr: '%s'."
+                % (self.name, cmd, exit_code, stderr))
+        ssh_remote_folder = stdout.split('\n')[0]
+
+        # Copy the input file(s) to remote directory.
+        for local_path, remote_path in list(app.inputs.items()):
             remote_path = os.path.join(ssh_remote_folder, remote_path)
             remote_parent = os.path.dirname(remote_path)
             try:
@@ -393,7 +434,7 @@ class BatchSystem(LRMS):
                 self.transport.put(local_path.path, remote_path)
                 # preserve execute permission on input files
                 if os.access(local_path.path, os.X_OK):
-                    self.transport.chmod(remote_path, 0755)
+                    self.transport.chmod(remote_path, 0o755)
             except:
                 log.critical(
                     "Copying input file '%s' to remote cluster '%s' failed",
@@ -404,17 +445,24 @@ class BatchSystem(LRMS):
             gc3libs.log.debug("Making remote path '%s' executable.",
                               app.arguments[0])
             self.transport.chmod(os.path.join(ssh_remote_folder,
-                                              app.arguments[0]), 0755)
+                                              app.arguments[0]), 0o755)
+
+        # if STDOUT/STDERR should be saved in a directory, ensure it
+        # exists (see Issue 495 for details)
+        for dest in (app.stdout, app.stderr):
+            if dest:
+                destdir = os.path.dirname(dest)
+                if destdir:
+                    self.transport.makedirs(
+                        posixpath.join(ssh_remote_folder, destdir))
 
         try:
             sub_cmd, aux_script = self._submit_command(app)
             if aux_script != '':
                 # create temporary script name
-                # XXX: The `uuid` module is available from Py 2.5 onwards
-                script_filename = ('./script.%x.sh'
-                                   % random.randint(0, sys.maxint))
+                script_filename = ('./script.%s.sh' % uuid.uuid4())
                 # save script to a temporary file and submit that one instead
-                local_script_file = tempfile.NamedTemporaryFile()
+                local_script_file = tempfile.NamedTemporaryFile(mode='wt')
                 local_script_file.write('#!/bin/sh\n')
                 # Add preamble file
                 prologue = self.get_prologue_script(app)
@@ -435,7 +483,7 @@ class BatchSystem(LRMS):
                     os.path.join(ssh_remote_folder, script_filename))
                 # set execution mode on remote script
                 self.transport.chmod(
-                    os.path.join(ssh_remote_folder, script_filename), 0755)
+                    os.path.join(ssh_remote_folder, script_filename), 0o755)
                 # cleanup
                 local_script_file.close()
                 if os.path.exists(local_script_file.name):
@@ -447,13 +495,13 @@ class BatchSystem(LRMS):
 
             # Submit it
             exit_code, stdout, stderr = self.transport.execute_command(
-                "/bin/sh -c 'cd %s && %s %s'"
-                % (ssh_remote_folder, sub_cmd, script_filename))
+                "/bin/sh -c %s" % sh_quote_safe('cd %s && %s %s' % (
+                    ssh_remote_folder, sub_cmd, script_filename)))
 
             if exit_code != 0:
                 raise gc3libs.exceptions.LRMSError(
-                    "Failed executing command 'cd %s && %s %s' on resource '%s';"
-                    " exit code: %d, stderr: '%s'."
+                    "Failed executing command 'cd %s && %s %s' on resource"
+                    " '%s'; exit code: %d, stderr: '%s'."
                     % (ssh_remote_folder, sub_cmd, script_filename,
                        self.name, exit_code, stderr))
 
@@ -499,128 +547,140 @@ class BatchSystem(LRMS):
             raise
 
 
-    def __do_acct(self, job, cmd, parse):
-        """Run `cmd` to get accounting information and update `job` state accordingly."""
+    def __run_command_and_parse_output(self, cmd, parser, kind='accounting'):
+        log.debug("Checking remote job %s info with `%s` ...", kind, cmd)
         exit_code, stdout, stderr = self.transport.execute_command(cmd)
         if exit_code == 0:
-            jobstatus = parse(stdout)
-            job.update(jobstatus)
-            if 'exitcode' in jobstatus:
-                job.returncode = int(jobstatus['exitcode'])
-                job.state = Run.State.TERMINATING
-            return job.state
+            return parser(stdout, stderr)
         else:
             raise gc3libs.exceptions.AuxiliaryCommandError(
-                "Failed running accounting command `%s`:"
+                "Failed running %s command `%s`:"
                 " exit code: %d, stderr: '%s'"
-                % (cmd, exit_code, stderr),
+                % (kind, cmd, exit_code, stderr),
                 do_log=True)
 
 
     @same_docstring_as(LRMS.update_job_state)
     @LRMS.authenticated
     def update_job_state(self, app):
+        job = app.execution
         try:
-            job = app.execution
             job.lrms_jobid
-        except AttributeError, ex:
+        except AttributeError as ex:
             # `job` has no `lrms_jobid`: object is invalid
             raise gc3libs.exceptions.InvalidArgument(
-                "Job object is invalid: %s" % str(ex))
+                "Job object is invalid: {ex}".format(ex=ex))
 
+        self.transport.connect()
+
+        cmd = self._stat_command(job)
         try:
-            self.transport.connect()
-            cmd = self._stat_command(job)
-            log.debug("Checking remote job status with '%s' ..." % cmd)
-            exit_code, stdout, stderr = self.transport.execute_command(cmd)
-            if exit_code == 0:
-                jobstatus = self._parse_stat_output(stdout)
-                job.update(jobstatus)
+            state, termstatus = self.__run_command_and_parse_output(
+                cmd, self._parse_stat_output, 'status')
+            if state != Run.State.TERMINATING:
+                # no need to go further and parse acct info; also,
+                # exit status is not relevant in this case
+                job.state = state
+                log.debug("Task %s state set to %s", app, state)
+                return state
+        except gc3libs.exceptions.AuxiliaryCommandError:
+            # use the special state value ``None`` to signal that
+            # the "status" command failed, we might need this
+            # after the "acct" command has run
+            state, termstatus = None, None
+        assert state is None or state == Run.State.TERMINATING
+        log.debug("Job status command gave state `%s`"
+                  " and termination status `%s` for task %s",
+                  state, termstatus, app)
 
-                job.state = jobstatus.get('state', Run.State.UNKNOWN)
-                if job.state == Run.State.UNKNOWN:
-                    log.warning(
-                        "Unknown batch job status,"
-                        " setting GC3Pie job state to `UNKNOWN`")
-
-                if 'exit_status' in jobstatus:
-                    job.exitcode = int(jobstatus['exit_status'])
-                    # XXX: we should set the `signal` part accordingly
-                    job.signal = 0
-
-                # SLURM's `squeue` command exits with code 0 if the
-                # job ID exists in the database (i.e., a job with that
-                # ID has been run) but prints no output.  In this
-                # case, we need to continue and examine the accounting
-                # command output to get the termination status etc.
-                if job.state != Run.State.TERMINATING:
-                    return job.state
-            else:
-                log.error(
-                    "Failed while running the `qstat`/`bjobs` command."
-                     " exit code: %d, stderr: '%s'" % (exit_code, stderr))
-
-            # In some batch systems, jobs disappear from qstat
-            # output as soon as they are finished. In these cases,
-            # we have to check some *accounting* command to check
-            # the exit status.
-            cmd = self._acct_command(job)
-            if cmd:
-                log.debug(
-                    "The job status command returned no information;"
-                    " trying with '%s' instead ...", cmd)
-                try:
-                    return self.__do_acct(job, cmd, self._parse_acct_output)
-                except gc3libs.exceptions.AuxiliaryCommandError:
-                    # This is used to distinguish between a standard
-                    # Torque installation and a PBSPro where `tracejob`
-                    # does not work but if `job_history_enable=True`,
-                    # then we can actually access information about
-                    # finished jobs with `qstat -x -f`.
-                    try:
-                        cmd = self._secondary_acct_command(job)
-                        if cmd:
-                            log.debug(
-                                "The primary job accounting command returned no"
-                                " information; trying with '%s' instead...", cmd)
-                            return self.__do_acct(job, cmd, self._parse_secondary_acct_output)
-                    except (gc3libs.exceptions.AuxiliaryCommandError,
-                            NotImplementedError):
-                        # ignore error -- there is nothing we can do
-                        pass
-
-            # No *stat command and no *acct command returned
-            # correctly.
+        # In some batch systems, jobs disappear from qstat
+        # output as soon as they are finished. In these cases,
+        # we have to check some *accounting* command to check
+        # the exit status.
+        acctinfo = {}
+        for cmd_fn, parse_fn in [
+                # this is the regular sacct/qacct/bjobs command
+                (self._acct_command, self._parse_acct_output),
+                # This is used to distinguish between a standard
+                # Torque installation and a PBSPro where `tracejob`
+                # does not work but if `job_history_enable=True`,
+                # then we can actually access information about
+                # finished jobs with `qstat -x -f`.
+                (self._secondary_acct_command, self._parse_secondary_acct_output),
+        ]:
+            cmd = cmd_fn(job)
+            # `._secondary_acct_command` returns ``None`` if no
+            # "secondary" accouting method is defined -- skip to next
+            # iteration, if any
+            if cmd is None:
+                continue
             try:
-                if (time.time() - job.stat_failed_at) > self.accounting_delay:
-                    # accounting info should be there, if it's not
-                    # then job is definitely lost
-                    log.critical(
-                        "Failed executing remote command: '%s';"
-                        "exit status %d", cmd, exit_code)
-                    log.debug(
-                        "  remote command returned stdout: '%s'", stdout)
-                    log.debug(
-                        "  remote command returned stderr: '%s'", stderr)
-                    raise gc3libs.exceptions.LRMSError(
-                        "Failed executing remote command: '%s'; exit status %d"
-                        % (cmd, exit_code))
-                else:
-                    # do nothing, let's try later...
-                    return job.state
+                acctinfo = self.__run_command_and_parse_output(
+                    cmd, parse_fn, 'accounting')
+                # use info from the first acct command that succeeds
+                if acctinfo:
+                    log.debug("Gathered accounting info %r for task %s",
+                              acctinfo, app)
+                    break
+            except gc3libs.exceptions.AuxiliaryCommandError:
+                log.debug("Accounting command `%s` failed.", cmd)
+                # try next one
+                pass
+            except gc3libs.exceptions.UnexpectedJobState as ex:
+                log.debug(
+                    "Unexpected output from accounting command `%s`: %s.",
+                    cmd, ex)
+                # try next one
+                pass
+
+        # if no termination status is known and the acct
+        # command provided one, use it
+        if 'termstatus' in acctinfo:
+            # if we have a termination status, then the job has terminated
+            state = Run.State.TERMINATING
+            if termstatus is None:
+                termstatus = acctinfo['termstatus']
+            else:
+                # this should not happen!  but one never knows how new
+                # versions of the software may break old habits and
+                # parsing rules, so better fail loudly here so we get
+                # a bug report and a chance to fix...
+                assert termstatus == acctinfo['termstatus'], (
+                    "Status and accounting commands disagree"
+                    " on job termination status: {1} vs {2}"
+                    .format(termstatus, acctinfo['termstatus'])
+                )
+
+        if termstatus is None:
+            # No *stat command and no *acct command returned correctly.
+            try:
+                job.stat_failed_at
             except AttributeError:
                 # this is the first time `qstat` fails, record a
                 # timestamp and retry later
                 job.stat_failed_at = time.time()
+                return job.state
 
-        except Exception, ex:
-            log.error("Error in querying Batch resource '%s': %s: %s",
-                      self.name, ex.__class__.__name__, str(ex))
-            raise
-        # If we reach this point it means that we don't actually know
-        # the current state of the job.
-        job.state = Run.State.UNKNOWN
-        return job.state
+            if (time.time() - job.stat_failed_at) <= self.accounting_delay:
+                # do nothing, let's try later...
+                return job.state
+            else:
+                # accounting info should be there, if it's not
+                # then job is definitely lost
+                job.state = Run.State.UNKNOWN
+                raise gc3libs.exceptions.LRMSError(
+                    "Could not retrieve status information for task {app}"
+                    .format(app=app))
+
+        # if we got to this point the job is in TERMINATING state
+        # and we know at least the termination status
+        assert state == Run.State.TERMINATING
+
+        job.state = state
+        job.returncode = termstatus
+        job.update(acctinfo)
+
+        return state
 
     @same_docstring_as(LRMS.peek)
     @LRMS.authenticated
@@ -631,7 +691,7 @@ class BatchSystem(LRMS):
             " passed to `PbsLrms.peek`."
 
         if size is None:
-            size = sys.maxint
+            size = sys.maxsize
 
         _filename_mapping = generic_filename_mapping(
             job.lrms_jobname, job.lrms_jobid, remote_filename)
@@ -644,7 +704,7 @@ class BatchSystem(LRMS):
                 _remote_filename, mode='r', bufsize=-1)
             remote_handler.seek(offset)
             data = remote_handler.read(size)
-        except Exception, ex:
+        except Exception as ex:
             log.error("Could not read remote file '%s': %s: %s",
                       _remote_filename, ex.__class__.__name__, str(ex))
 
@@ -662,7 +722,7 @@ class BatchSystem(LRMS):
         Supported protocols: file
         """
         for url in data_file_list:
-            if not url.scheme in ['file']:
+            if url.scheme not in ['file']:
                 return False
         return True
 
@@ -701,15 +761,15 @@ class BatchSystem(LRMS):
         try:
             self.transport.connect()
             self.transport.remove_tree(job.ssh_remote_folder)
-        except:
+        except Exception as err:
             log.warning("Failed removing remote folder '%s': %s: %s",
-                        job.ssh_remote_folder, sys.exc_info()[0],
-                        sys.exc_info()[1])
+                        job.ssh_remote_folder, err.__class__, err)
         return
 
     @same_docstring_as(LRMS.get_results)
     @LRMS.authenticated
-    def get_results(self, app, download_dir, overwrite=False, changed_only=True):
+    def get_results(self, app, download_dir,
+                    overwrite=False, changed_only=True):
         if app.output_base_url is not None:
             raise gc3libs.exceptions.UnrecoverableDataStagingError(
                 "Retrieval of output files to non-local destinations"
@@ -723,7 +783,7 @@ class BatchSystem(LRMS):
             # `Application.outputs` list to expand wildcards and
             # directory references.
             stageout = list()
-            for remote_relpath, local_url in app.outputs.iteritems():
+            for remote_relpath, local_url in app.outputs.items():
                 local_relpath = local_url.path
                 if remote_relpath == gc3libs.ANY_OUTPUT:
                     remote_relpath = ''
