@@ -2,7 +2,7 @@
 #
 """
 """
-# Copyright (C) 2011-2014, GC3, University of Zurich. All rights reserved.
+# Copyright (C) 2011-2014, 2016  University of Zurich. All rights reserved.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -15,52 +15,58 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+from __future__ import absolute_import, print_function, unicode_literals
+from builtins import object
 __docformat__ = 'reStructuredText'
-__version__ = '$Revision$'
 
 import datetime
 import os
 import shutil
-import sys
 import tempfile
 
 import gc3libs
 import gc3libs.core
 import gc3libs.config
 from gc3libs.backends.lsf import LsfLrms
-from gc3libs.quantity import Duration, hours, minutes, seconds, Memory, GB, MB, kB
-
-from nose.tools import assert_equal, raises
+from gc3libs.quantity import Duration, hours, Memory, GB
 
 _datetime_date = None
 
 files_to_remove = []
 
+
 def setUpModule():
-    """Mock the `datetime.date.today()` outcome in order to make the LSF parsing independent from the testing date."""
+    """Mock the `datetime.date.today()` outcome in order to make the LSF
+    parsing independent from the testing date."""
     # save the original `datetime.date` to restore it in `tearDownModule()`
     import datetime
     global _datetime_date
     _datetime_date = datetime.date
     # mock features of `datetime.date.today()` that are actually used
     # in `LsfLrms._parse_date()`
+
     class MockDate(object):
+
         def __init__(self, real):
             self.__date = real
+
         def __getattr__(self, name):
             return getattr(self.__date, name)
+
         def __call__(self, *args, **kwargs):
             return self.__date(*args, **kwargs)
     datetime.date = MockDate(datetime.date)
+
     class Today(object):
+
         def __init__(self):
             self.year = 2012
             self.month = 12
     datetime.date.today = Today
 
+
 def tearDownModule():
     # restore the original `datetime.date`
     global _datetime_date
@@ -72,6 +78,7 @@ def tearDownModule():
         else:
             os.remove(fname)
 
+
 def test_get_command():
     (fd, tmpfile) = tempfile.mkstemp()
     files_to_remove.append(tmpfile)
@@ -106,12 +113,12 @@ lsf_continuation_line_prefix_length = 12
     cfg.merge_file(tmpfile)
     b = cfg.make_resources()['example']
 
-    assert_equal(b.bsub, ['/usr/local/bin/bsub', '-R', 'lustre'])
+    assert b.bsub == ['/usr/local/bin/bsub', '-R', 'lustre']
 
-    assert_equal(b._bjobs,   '/usr/local/bin/bjobs')
-    assert_equal(b._lshosts, '/usr/local/sbin/lshosts')
+    assert b._bjobs == '/usr/local/bin/bjobs'
+    assert b._lshosts == '/usr/local/sbin/lshosts'
 
-    assert_equal(b._CONTINUATION_LINE_START, 12 * ' ')
+    assert b._CONTINUATION_LINE_START == 12 * ' '
 
 
 def test_bjobs_output_done1():
@@ -119,12 +126,12 @@ def test_bjobs_output_done1():
                   architecture=gc3libs.Run.Arch.X86_64,
                   max_cores=1,
                   max_cores_per_job=1,
-                  max_memory_per_core=1*GB,
-                  max_walltime=1*hours,
-                  auth=None, # ignored if `transport` is `local`
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
                   frontend='localhost',
                   transport='local')
-    jobstatus = lsf._parse_stat_output("""
+    bjobs_output = """
 Job <131851>, Job Name <ChromaExtractShort>, User <wwolski>, Project <default>,
                      Status <DONE>, Queue <pub.8h>, Job Priority <50>, Command
                      <ChromatogramExtractor -in /cluster/scratch/malars/openswa
@@ -149,6 +156,9 @@ Tue Jul 24 10:04:19: Started on <a6122>, Execution Home </cluster/home/biol/wwo
                      obs/lrms_job.YNZmU17755/.>;
 Tue Jul 24 10:05:45: Done successfully. The CPU time used is 2.1 seconds.
 
+ MEMORY USAGE:
+ MAX MEM: 41 Mbytes;  AVG MEM: 41 Mbytes
+
  SCHEDULING PARAMETERS:
            r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
  loadSched   -     -     -     -       -     -    -     -  1000M     -      -
@@ -157,9 +167,65 @@ Tue Jul 24 10:05:45: Done successfully. The CPU time used is 2.1 seconds.
           scratch      xs       s       m       l      xl      sp
  loadSched 4000.0      -       -       -       -       -       -
  loadStop      -       -       -       -       -       -       -
-""")
-    assert_equal(jobstatus.state, gc3libs.Run.State.TERMINATING)
-    assert_equal(jobstatus.exit_status, 0)
+"""
+    jobstatus = lsf._parse_stat_output(bjobs_output, '')
+    assert jobstatus.state == gc3libs.Run.State.TERMINATING
+    assert jobstatus.termstatus == (0, 0)
+
+
+def test_bjobs_output_for_accounting():
+    lsf = LsfLrms(name='test',
+                  architecture=gc3libs.Run.Arch.X86_64,
+                  max_cores=1,
+                  max_cores_per_job=1,
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
+                  frontend='localhost',
+                  transport='local')
+    bjobs_output = """
+Job <131851>, Job Name <ChromaExtractShort>, User <wwolski>, Project <default>,
+                     Status <DONE>, Queue <pub.8h>, Job Priority <50>, Command
+                     <ChromatogramExtractor -in /cluster/scratch/malars/openswa
+                     th/data/AQUA_fixed_water/split_napedro_L120224_001_SW-400A
+                     QUA_no_background_2ul_dilution_10/split_napedro_L120224_00
+                     1_SW-400AQUA_no_background_2ul_dilution_10_28.mzML.gz -tr
+                     /cluster/scratch/malars/openswath/assays/iRT/DIA_iRT.TraML
+                      -out split_napedro_L120224_001_SW-400AQUA_no_background_2
+                     ul_dilution_10_28._rtnorm.chrom.mzML -is_swath -min_upper_
+                     edge_dist 1 -threads 2>, Share group charged </lsf_biol_al
+                     l/lsf_biol_other/wwolski>
+Tue Jul 24 10:03:15: Submitted from host <brutus3>, CWD <$HOME/.gc3pie_jobs/lrm
+                     s_job.YNZmU17755/.>, Output File <lsf.o%J>, Requested Reso
+                     urces <select[mem<70000 && lustre] order[-ut] rusage[mem=1
+                     000,m=1]>, Login Shell </bin/sh>, Specified Hosts <thin+9>
+                     , <single+8>, <smp16+6>, <smp24+5>, <smp48+4>;
+
+ RUNLIMIT
+ 480.0 min of a6122
+Tue Jul 24 10:04:19: Started on <a6122>, Execution Home </cluster/home/biol/wwo
+                     lski>, Execution CWD </cluster/home/biol/wwolski/.gc3pie_j
+                     obs/lrms_job.YNZmU17755/.>;
+Tue Jul 24 10:05:45: Done successfully. The CPU time used is 2.1 seconds.
+
+ MEMORY USAGE:
+ MAX MEM: 41 Mbytes;  AVG MEM: 41 Mbytes
+
+ SCHEDULING PARAMETERS:
+           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
+ loadSched   -     -     -     -       -     -    -     -  1000M     -      -
+ loadStop    -     -     -     -       -     -    -     -     -      -      -
+
+          scratch      xs       s       m       l      xl      sp
+ loadSched 4000.0      -       -       -       -       -       -
+ loadStop      -       -       -       -       -       -       -
+"""
+
+    # Also parse the output of jobs to get accounting information
+    acct = lsf._parse_acct_output(bjobs_output, '')
+    assert acct['duration'] == Duration('86s')
+    assert acct['used_cpu_time'] == Duration('2.1s')
+    assert acct['max_used_memory'] == Memory('41MB')
 
 
 def test_bjobs_output_done2():
@@ -167,9 +233,9 @@ def test_bjobs_output_done2():
                   architecture=gc3libs.Run.Arch.X86_64,
                   max_cores=1,
                   max_cores_per_job=1,
-                  max_memory_per_core=1*GB,
-                  max_walltime=1*hours,
-                  auth=None, # ignored if `transport` is `local`
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
                   frontend='localhost',
                   transport='local')
     jobstatus = lsf._parse_stat_output("""
@@ -268,9 +334,11 @@ Mon Jul 30 15:12:56: Done successfully. The CPU time used is 1.7 seconds.
           scratch      xs       s       m       l      xl      sp
  loadSched20000.0      -       -       -       -       -       -
  loadStop      -       -       -       -       -       -       -
-""")
-    assert_equal(jobstatus.state, gc3libs.Run.State.TERMINATING)
-    assert_equal(jobstatus.exit_status, 0)
+""",
+    # STDERR
+    '')
+    assert jobstatus.state == gc3libs.Run.State.TERMINATING
+    assert jobstatus.termstatus == (0, 0)
 
 
 def test_bjobs_output_done3():
@@ -278,9 +346,9 @@ def test_bjobs_output_done3():
                   architecture=gc3libs.Run.Arch.X86_64,
                   max_cores=1,
                   max_cores_per_job=1,
-                  max_memory_per_core=1*GB,
-                  max_walltime=1*hours,
-                  auth=None, # ignored if `transport` is `local`
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
                   frontend='localhost',
                   transport='local')
     jobstatus = lsf._parse_stat_output("""
@@ -311,9 +379,31 @@ Mon Aug  4 12:28:51 2014: Completed <exit>.
  RESOURCE REQUIREMENT DETAILS:
  Combined: select[type == local] order[r15s:pg] rusage[mem=2000.00]
  Effective: select[type == local] order[r15s:pg] rusage[mem=2000.00]
-""")
-    assert_equal(jobstatus.state, gc3libs.Run.State.TERMINATING)
-    assert_equal(jobstatus.exit_status, 127)
+    """,
+    # STDERR
+    '')
+    assert jobstatus.state == gc3libs.Run.State.TERMINATING
+    assert jobstatus.termstatus == (0, 127)
+
+
+def test_bjobs_output_done_long_ago():
+    """Test parsing `bjobs -l` output for a job that was removed from `mbatchd` core memory"""
+    lsf = LsfLrms(name='test',
+                  architecture=gc3libs.Run.Arch.X86_64,
+                  max_cores=1,
+                  max_cores_per_job=1,
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
+                  frontend='localhost',
+                  transport='local')
+    jobstatus = lsf._parse_stat_output(
+        # empty STDOUT
+        '',
+        # STDERR
+        'Job <943186> is not found')
+    assert jobstatus.state == gc3libs.Run.State.TERMINATING
+    assert jobstatus.termstatus == None
 
 
 def test_bjobs_output_exit_nonzero():
@@ -321,9 +411,9 @@ def test_bjobs_output_exit_nonzero():
                   architecture=gc3libs.Run.Arch.X86_64,
                   max_cores=1,
                   max_cores_per_job=1,
-                  max_memory_per_core=1*GB,
-                  max_walltime=1*hours,
-                  auth=None, # ignored if `transport` is `local`
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
                   frontend='localhost',
                   transport='local')
     jobstatus = lsf._parse_stat_output("""
@@ -348,24 +438,25 @@ Tue Jul 24 10:26:53: Completed <exit>.
           scratch      xs       s       m       l      xl      sp
  loadSched20000.0      -       -       -       -       -       -
  loadStop      -       -       -       -       -       -       -
-""")
-    assert_equal(jobstatus.state, gc3libs.Run.State.TERMINATING)
-    assert_equal(jobstatus.exit_status, 42)
+    """,
+    # STDERR
+    '')
+    assert jobstatus.state == gc3libs.Run.State.TERMINATING
+    assert jobstatus.termstatus == (0, 42)
 
 
-@raises(AssertionError)
 def test_bjobs_incorrect_prefix_length():
     lsf = LsfLrms(name='test',
                   architecture=gc3libs.Run.Arch.X86_64,
                   max_cores=1,
                   max_cores_per_job=1,
-                  max_memory_per_core=1*GB,
-                  max_walltime=1*hours,
-                  auth=None, # ignored if `transport` is `local`
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
                   frontend='localhost',
                   transport='local',
                   lsf_continuation_line_prefix_length=7)
-    jobstatus = lsf._parse_stat_output("""
+    stat_result = lsf._parse_stat_output("""
 Job <2073>, Job Name <GRunApplication.0>, User <markmon>, Project <default>, St
                           atus <EXIT>, Queue <normal>, Command <sh -c inputfile
                           .txt>
@@ -393,8 +484,11 @@ Mon Aug  4 12:28:51 2014: Completed <exit>.
  RESOURCE REQUIREMENT DETAILS:
  Combined: select[type == local] order[r15s:pg] rusage[mem=2000.00]
  Effective: select[type == local] order[r15s:pg] rusage[mem=2000.00]
-""")
-    assert 'state' not in jobstatus
+""",
+    # STDERR
+    '')
+    assert stat_result.state == gc3libs.Run.State.UNKNOWN
+    assert stat_result.termstatus == None
 
 
 def test_bjobs_correct_explicit_prefix_length():
@@ -402,13 +496,13 @@ def test_bjobs_correct_explicit_prefix_length():
                   architecture=gc3libs.Run.Arch.X86_64,
                   max_cores=1,
                   max_cores_per_job=1,
-                  max_memory_per_core=1*GB,
-                  max_walltime=1*hours,
-                  auth=None, # ignored if `transport` is `local`
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
                   frontend='localhost',
                   transport='local',
                   lsf_continuation_line_prefix_length=26)
-    jobstatus = lsf._parse_stat_output("""
+    stat_result = lsf._parse_stat_output("""
 Job <2073>, Job Name <GRunApplication.0>, User <markmon>, Project <default>, St
                           atus <EXIT>, Queue <normal>, Command <sh -c inputfile
                           .txt>
@@ -436,16 +530,27 @@ Mon Aug  4 12:28:51 2014: Completed <exit>.
  RESOURCE REQUIREMENT DETAILS:
  Combined: select[type == local] order[r15s:pg] rusage[mem=2000.00]
  Effective: select[type == local] order[r15s:pg] rusage[mem=2000.00]
-""")
-    assert 'state' in jobstatus
-    assert_equal(jobstatus.state, gc3libs.Run.State.TERMINATING)
-    assert_equal(jobstatus.exit_status, 127)
+""",
+    # STDERR
+    '')
+    assert stat_result.state == gc3libs.Run.State.TERMINATING
+    assert stat_result.termstatus == (0, 127)
 
 
 def test_bacct_done0():
     """Test parsing accounting information of a <sleep 300> job."""
     # gotten with `bacct -l "jobid"`
-    acct = LsfLrms._parse_acct_output("""
+    lsf = LsfLrms(name='test',
+                  architecture=gc3libs.Run.Arch.X86_64,
+                  max_cores=1,
+                  max_cores_per_job=1,
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
+                  frontend='localhost',
+                  transport='local',
+                  bacct='bacct')
+    acct = lsf._parse_acct_output("""
 Accounting information about jobs that are:
   - submitted by all users.
   - accounted on all projects.
@@ -481,20 +586,35 @@ SUMMARY:      ( time unit: second )
  Average hog factor of a job:  0.00 ( cpu time / turnaround time )
  Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00
 
-    """)
-    assert_equal(acct['duration'], Duration('67s'))
-    assert_equal(acct['used_cpu_time'], Duration('0.08s'))
-    assert_equal(acct['max_used_memory'], Memory('227MB'))
+    """,
+    # STDERR
+    '')
+    assert acct['duration'] == Duration('67s')
+    assert acct['used_cpu_time'] == Duration('0.08s')
+    assert acct['max_used_memory'] == Memory('227MB')
     # timestamps
     year = datetime.date.today().year
-    assert_equal(acct['lsf_submission_time'], datetime.datetime(year, 10, 8, 17, 7, 54))
-    assert_equal(acct['lsf_start_time'],      datetime.datetime(year, 10, 8, 17, 8, 44))
-    assert_equal(acct['lsf_completion_time'], datetime.datetime(year, 10, 8, 17, 9, 51))
+    assert (acct['lsf_submission_time'] ==
+                 datetime.datetime(year, 10, 8, 17, 7, 54))
+    assert (acct['lsf_start_time'] ==
+                 datetime.datetime(year, 10, 8, 17, 8, 44))
+    assert (acct['lsf_completion_time'] ==
+                 datetime.datetime(year, 10, 8, 17, 9, 51))
 
 
 def test_bacct_done1():
     """Test parsing `bacct -l` output for a not-so-trivial job."""
-    acct = LsfLrms._parse_acct_output("""
+    lsf = LsfLrms(name='test',
+                  architecture=gc3libs.Run.Arch.X86_64,
+                  max_cores=1,
+                  max_cores_per_job=1,
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
+                  frontend='localhost',
+                  transport='local',
+                  bacct='bacct')
+    acct = lsf._parse_acct_output("""
 Accounting information about jobs that are:
   - submitted by all users.
   - accounted on all projects.
@@ -529,20 +649,35 @@ SUMMARY:      ( time unit: second )
  Maximum turnaround time:        73      Minimum turnaround time:        73
  Average hog factor of a job:  0.00 ( cpu time / turnaround time )
  Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00
-    """)
-    assert_equal(acct['duration'], Duration('6s'))
-    assert_equal(acct['used_cpu_time'], Duration('0.04s'))
-    assert_equal(acct['max_used_memory'], Memory('37MB'))
+    """,
+    # STDERR
+    '')
+    assert acct['duration'] == Duration('6s')
+    assert acct['used_cpu_time'] == Duration('0.04s')
+    assert acct['max_used_memory'] == Memory('37MB')
     # timestamps
     year = datetime.date.today().year
-    assert_equal(acct['lsf_submission_time'], datetime.datetime(year, 10, 8, 17,  8, 54))
-    assert_equal(acct['lsf_start_time'],      datetime.datetime(year, 10, 8, 17, 10,  1))
-    assert_equal(acct['lsf_completion_time'], datetime.datetime(year, 10, 8, 17, 10,  7))
+    assert (acct['lsf_submission_time'] ==
+                 datetime.datetime(year, 10, 8, 17, 8, 54))
+    assert (acct['lsf_start_time'] ==
+                 datetime.datetime(year, 10, 8, 17, 10, 1))
+    assert (acct['lsf_completion_time'] ==
+                 datetime.datetime(year, 10, 8, 17, 10, 7))
 
 
 def test_bacct_killed():
     """Test parsing `bacct -l` output for a canceled job."""
-    acct = LsfLrms._parse_acct_output("""
+    lsf = LsfLrms(name='test',
+                  architecture=gc3libs.Run.Arch.X86_64,
+                  max_cores=1,
+                  max_cores_per_job=1,
+                  max_memory_per_core=1 * GB,
+                  max_walltime=1 * hours,
+                  auth=None,  # ignored if `transport` is `local`
+                  frontend='localhost',
+                  transport='local',
+                  bacct='bacct')
+    acct = lsf._parse_acct_output("""
 Accounting information about jobs that are:
   - submitted by all users.
   - accounted on all projects.
@@ -577,36 +712,40 @@ SUMMARY:      ( time unit: second )
  Maximum turnaround time:       115      Minimum turnaround time:       115
  Average hog factor of a job:  0.00 ( cpu time / turnaround time )
  Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00
-""")
-    assert_equal(acct['duration'], Duration('55s'))
-    assert_equal(acct['used_cpu_time'], Duration('0.04s'))
-    assert_equal(acct['max_used_memory'], Memory('35MB'))
+""",
+    # STDERR
+    '')
+    assert acct['duration'] == Duration('55s')
+    assert acct['used_cpu_time'] == Duration('0.04s')
+    assert acct['max_used_memory'] == Memory('35MB')
     # timestamps
     year = datetime.date.today().year
-    assert_equal(acct['lsf_submission_time'], datetime.datetime(year, 10, 5, 17, 49, 35))
-    assert_equal(acct['lsf_start_time'],      datetime.datetime(year, 10, 5, 17, 50, 35))
-    assert_equal(acct['lsf_completion_time'], datetime.datetime(year, 10, 5, 17, 51, 30))
-
+    assert (acct['lsf_submission_time'] ==
+                 datetime.datetime(year, 10, 5, 17, 49, 35))
+    assert (acct['lsf_start_time'] ==
+                 datetime.datetime(year, 10, 5, 17, 50, 35))
+    assert (acct['lsf_completion_time'] ==
+                 datetime.datetime(year, 10, 5, 17, 51, 30))
 
 
 # LSF incorporates resource usage information in a job's output;
 # the job's output is a copy of the email that the LSF system
 # sends to the user that submitted a job.
-################################################################################################
+###############################################################################
 # Sender: LSF System <lsfadmin@cpt086>
 # Subject: Job 943186: <md5sum XXX/x.sh> Done
 #
-# Job <md5sum XXX/x.sh> was submitted from host <frt> by user <rmurri> in cluster <prdclst>.
-# Job was executed on host(s) <cpt086>, in queue <normal>, as user <rmurri> in cluster <prdclst>.
-# </home/rmurri> was used as the home directory.
-# </home/rmurri> was used as the working directory.
-# Started at Mon Oct  8 17:23:30 2012
-# Results reported at Mon Oct  8 17:23:31 2012
+# Job <md5sum XXX/x.sh> was submitted from host <frt> by user <rmurri>
+# in cluster <prdclst>.  Job was executed on host(s) <cpt086>, in
+# queue <normal>, as user <rmurri> in cluster <prdclst>.
+# </home/rmurri> was used as the home directory.  </home/rmurri> was
+# used as the working directory.  Started at Mon Oct 8 17:23:30 2012
+# Results reported at Mon Oct 8 17:23:31 2012
 #
 # Your job looked like:
 #
 # ------------------------------------------------------------
-# # LSBATCH: User input
+# LSBATCH: User input
 # md5sum XXX/x.sh
 # ------------------------------------------------------------
 #
@@ -624,9 +763,9 @@ SUMMARY:      ( time unit: second )
 # The output (if any) follows:
 #
 # b7f9d9c86469aa8b57c19af14bf80af9  XXX/x.sh
-################################################################################################
+###############################################################################
 
 
 if __name__ == "__main__":
-    import nose
-    nose.runmodule()
+    import pytest
+    pytest.main(["-v", __file__])
