@@ -6,7 +6,7 @@ execute commands and copy/move files irrespective of whether the
 destination is the local computer or a remote front-end that we access
 via SSH.
 """
-# Copyright (C) 2009-2014 GC3, University of Zurich. All rights reserved.
+# Copyright (C) 2009-2019  University of Zurich. All rights reserved.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -19,20 +19,35 @@ via SSH.
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+from __future__ import absolute_import, print_function, unicode_literals
+from __future__ import division
+from builtins import str
+from past.utils import old_div
+from builtins import object
 __docformat__ = 'reStructuredText'
-__version__ = '2.2.3 version (SVN $Revision: 4001 $)'
 
 
+import platform
 import os
 import os.path
 import errno
 import shutil
 import getpass
+import shutil
 
-from gc3libs.utils import same_docstring_as, samefile
+try:
+    # Python 2
+    from types import StringTypes as string_types
+except ImportError:
+    # Python 3
+    string_types = (str,)
+
+
+import gc3libs.defaults
+from gc3libs.quantity import Memory, MiB
+from gc3libs.utils import same_docstring_as, samefile, to_str
 import gc3libs.exceptions
 
 
@@ -95,8 +110,7 @@ class Transport(object):
 
     def get(self, source, destination, ignore_nonexisting=False,
             overwrite=False, changed_only=True):
-        """
-        Copy remote `source` to local `destination`.
+        """Copy remote `source` to local `destination`.
 
         Permission bits are copied. Both `source` and `destination`
         are path names given as strings.
@@ -135,9 +149,15 @@ class Transport(object):
 
         :param str source: the file or directory to copy
         :param str destination: the destination file or directory
-        :param bool ignore_nonexisting: if `True`, no exceptions will be raised if `source` does not exist
+        :param bool ignore_nonexisting: if `True`, no exceptions will
+                                        be raised if `source` does not exist
         :param bool overwrite: if `True`, overwrite existing destination files
-        :param bool changed_only: if both this and `overwrite` are `True`, only overwrite those files such that the source is newer or different in size than the destination.
+        :param bool changed_only: if both this and `overwrite` are
+                                  `True`, only overwrite those files
+                                  such that the source is newer or
+                                  different in size than the
+                                  destination.
+
         """
         try:
             if self.isdir(source):
@@ -162,7 +182,7 @@ class Transport(object):
                         sst = self.stat(source)
                         dst = os.stat(destination)
                         if (sst.st_size == dst.st_size
-                            and sst.st_mtime <= dst.st_mtime):
+                                and sst.st_mtime <= dst.st_mtime):
                             gc3libs.log.debug(
                                 "Transport.get(): Local file '%s'"
                                 " has same size and modification time as"
@@ -175,9 +195,11 @@ class Transport(object):
                 if not os.path.exists(parent):
                     os.makedirs(parent)
                 self._get_impl(source, destination)
-        except Exception, ex:
+        except Exception as ex:
             # IOError(errno=2) means the remote path is not existing
-            if (ignore_nonexisting and isinstance(ex, IOError) and ex.errno == 2):
+            if (ignore_nonexisting
+                    and isinstance(ex, IOError)
+                    and ex.errno == 2):
                 pass
             else:
                 raise gc3libs.exceptions.TransportError(
@@ -229,7 +251,7 @@ class Transport(object):
             "Abstract method `Transport.listdir()` called - "
             "this should have been defined in a derived class.")
 
-    def makedirs(self, path, mode=0777):
+    def makedirs(self, path, mode=0o777):
         """
         Recursive directory creation function. Makes all
         intermediate-level directories needed to contain the leaf
@@ -277,17 +299,13 @@ class Transport(object):
         try:
             destdir = os.path.dirname(destination)
             self.makedirs(destdir)
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not make directory '%s' on host '%s': %s: %s"
                 % (destdir, self.remote_frontend,
                    ex.__class__.__name__, str(ex)))
         try:
             if os.path.isdir(source):
-                # destination must be a directory
-                assert self.isdir(destination), \
-                    "Transport.put(): when source is a directory," \
-                    " destination must be a directory, too!"
                 # `source` is a directory, recursively descend it
                 self.makedirs(destination)
                 for entry in os.listdir(source):
@@ -310,7 +328,7 @@ class Transport(object):
                         sst = os.stat(source)
                         dst = self.stat(destination)
                         if (sst.st_size == dst.st_size
-                            and sst.st_mtime <= dst.st_mtime):
+                                and sst.st_mtime <= dst.st_mtime):
                             gc3libs.log.debug(
                                 "Tranport.put(): Remote file '%s'"
                                 " has same size and modification time as"
@@ -332,7 +350,7 @@ class Transport(object):
                         pass
                     else:
                         raise
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not upload '%s' to '%s' on host '%s': %s: %s"
                 % (source, destination, self.remote_frontend,
@@ -394,7 +412,6 @@ class Transport(object):
 # SSH Transport class
 #
 
-import sys
 import types
 
 import paramiko
@@ -404,44 +421,167 @@ import gc3libs
 
 class SshTransport(Transport):
 
-    def __init__(self, remote_frontend, port=gc3libs.Default.SSH_PORT,
-                 username=None, ignore_ssh_host_keys=False, keyfile=None,
-                 timeout=gc3libs.Default.SSH_CONNECT_TIMEOUT):
-        self.remote_frontend = remote_frontend
-        self.port = port
-        self.username = username
+    def __init__(self, remote_frontend,
+                 ignore_ssh_host_keys=False,
+                 ssh_config=None,
+                 username=None, port=None,
+                 keyfile=None, timeout=None,
+                 large_file_threshold=None,
+                 large_file_chunk_size=None,
+                 **extra_args):
+        """
+        Initialize an `SshTransport` object for operating on host `remote_frontend`.
 
+        Second optional argument `ignore_ssh_host_keys` instructs the
+        communication layer *not* to validate the SSH host key and
+        ignore the contents of SSH "known hosts" file.  While
+        insecure, this is the only way of dealing with cloud-based VMs
+        (where the host key is generated during VM creation).
+
+        Third optional argument `ssh_config` specifies the path to an
+        OpenSSH configuration file (see man page `ssh_config(5)` for
+        details).  If found in that file, any of the following options
+        override the GC3Pie built-in default:
+
+        * ``ConnectTimeout`` sets the maximum time GC3Pie will wait
+          for a connection to be established, before considering the
+          attempt failed.
+        * ``HostName`` overrides the host name given by `remote_frontend`
+        * ``IdentityFile`` sets the (private) key file to use for
+          authentication to the remote host.
+        * ``Port`` sets the TCP port to use for connections.
+        * ``ProxyCommand`` pipes all SSH I/O through the given command.
+        * ``User`` sets the user name to use when connecting.
+
+        Additional arguments ``user``, ``port``, ``keyfile``, and
+        ``timeout``, if given, override the above settings.
+
+        Finally, the two parameters `large_file_threshold` and
+        `large_file_chunk_size` control how file copy is being made:
+        if a file is larger than `large_file_threshold`, then it will
+        be transferred by sequentially copying chunks of
+        `large_file_chunk_size` bytes at a time; else, the entire file
+        will be requested at once, using many parallel block
+        transfers.  See `SshTransport.get()`:meth: for details.
+        """
         self.ssh = paramiko.SSHClient()
-        self.timeout = timeout
         self.ignore_ssh_host_keys = ignore_ssh_host_keys
         self.sftp = None
         self._is_open = False
         self.transport_channel = None
 
-        if keyfile is None:
-            # try to get the path to SSH public key file  from `~/.ssh/config`
-            try:
-                ssh_config = paramiko.SSHConfig()
-                config_filename = os.path.expanduser('~/.ssh/config')
-                config_file = open(config_filename)
-                ssh_config.parse(config_file)
-                # Check if we have an ssh configuration stanza for this host
-                hostconfig = ssh_config.lookup(self.remote_frontend)
-                self.keyfile = hostconfig.get('identityfile', None)
-                config_file.close()
-            except IOError, err:
-                if err.errno == 2:
-                    # "No such file or directory" -- ignore error
-                    self.keyfile = None
-                else:
-                    raise
+        # init connection params
+        self.username = None
+        self.keyfile = None
+        self.port = gc3libs.defaults.SSH_PORT
+        self.timeout = gc3libs.defaults.SSH_CONNECT_TIMEOUT
+        self.proxy_command = None
+
+        # use SSH options, if available
+        self._ssh_config = paramiko.SSHConfig()
+        config_filename = os.path.expanduser(ssh_config or gc3libs.defaults.SSH_CONFIG_FILE)
+        if os.path.exists(config_filename):
+            with open(config_filename, 'r') as config_file:
+                self._ssh_config.parse(config_file)
+        self.set_connection_params(remote_frontend, username, keyfile, port, timeout)
+
+        # SSH copy size params; convert to int for more efficiency at time of use
+        self.large_file_threshold = self._memory_to_bytes(
+            large_file_threshold or self._estimate_safe_buffer_size())
+        self.large_file_chunk_size = self._memory_to_bytes(
+            large_file_chunk_size or self._estimate_safe_buffer_size())
+
+        if __debug__ and extra_args:
+            gc3libs.log.debug(
+                "SshTransport: ignoring extra init arguments: %s",
+                ', '.join("{0}={1!r}".format(k, v)
+                          for k,v in extra_args.items()))
+
+    @staticmethod
+    def _estimate_safe_buffer_size():
+        """
+        Return estimate for max size of buffer for file transfers.
+
+        See `SshTransport._get_impl`:meth: for details of how this is
+        used and why it is needed.
+        """
+        avail_mem = gc3libs.utils.get_max_real_memory()
+        if avail_mem is not None:
+            # be sure to use no more than 50% of avail mem
+            # if we cannot determine number of processors
+            nproc = gc3libs.utils.get_num_processors() or 2
+            return (old_div(avail_mem, nproc))
         else:
-            assert type(keyfile) in types.StringTypes
+            # no clue how much memory is available, fall back to
+            # (hard-coded) 32MiB which should be safe on today's computers
+            return 32*MiB
+
+    @staticmethod
+    def _memory_to_bytes(qty):
+        try:
+            return qty.amount(Memory.B, conv=int)
+        except AttributeError:
+            pass
+        try:
+            return int(qty)
+        except (ValueError, TypeError):
+            raise TypeError(
+                "`gc3libs.quantity.Memory` or integer (count of bytes) expected,"
+                " instead gotten {0!r} {1}"
+                .format(qty, type(qty)))
+
+    def set_connection_params(self, hostname, username=None, keyfile=None,
+                               port=None, timeout=None):
+        """
+        Set remote host name and other parameters used for new connections.
+        Currently-established connections are not affected.
+
+        The host name is the only mandatory argument; any other parameter will
+        be read from the SSH configuration file (unless explicitly provided to
+        this function).
+        """
+        # check if we have an ssh configuration stanza for this host
+        ssh_options = self._ssh_config.lookup(hostname)
+
+        # merge SSH options from the SSH config file with parameters
+        # we were given in this method call
+        self.remote_frontend = ssh_options.get('hostname', hostname)
+
+        if username is None:
+            self.username = ssh_options.get('user', self.username)
+        else:
+            assert type(username) in string_types
+            self.username = username
+
+        if port is None:
+            self.port = int(ssh_options.get('port', self.port))
+        else:
+            self.port = int(port)
+
+        if keyfile is None:
+            self.keyfile = ssh_options.get('identityfile', self.keyfile)
+        else:
+            assert type(keyfile) in string_types
             self.keyfile = keyfile
 
+        if timeout is None:
+            self.timeout = float(ssh_options.get('connecttimeout', self.timeout))
+        else:
+            self.timeout = float(timeout)
+
+        # support for extra configuration options, not having a direct
+        # equivalent in the GC3Pie configuration file
+        self.proxy_command = ssh_options.get('proxycommand', None)
+
 
     @same_docstring_as(Transport.connect)
     def connect(self):
+        if not self.remote_frontend:
+            self._is_open = False
+            raise gc3libs.exceptions.TransportError(
+                "Cannot connect to remote host:"
+                " no host name/IP address known yet.")
+
         try:
             self.transport_channel = self.ssh.get_transport()
             if not self._is_open or self.transport_channel is None or \
@@ -452,26 +592,50 @@ class SshTransport(Transport):
                     # hosts" database file. This is needed for EC2
                     # backends in order to fix issue 389, but
                     # introduces a security risk in normal situations,
-                    # thus the check is enabled by default.
-                    self.ssh.load_system_host_keys()
+                    # thus the check is enabled by default. However,
+                    # Paramiko can fail to parse `~/.ssh/known_hosts`
+                    # (seen on MacOSX) and then raise an
+                    # `SSHException` which causes the whole block to
+                    # fail.  So, ignore any errors raised by this line
+                    # and hope for the best.
+                    try:
+                        self.ssh.load_system_host_keys()
+                    except paramiko.SSHException as err:
+                        gc3libs.log.warning(
+                            "Could not read 'known hosts' SSH keys (%s: %s)."
+                            " I'm ignoring the error and continuing anyway,"
+                            " but this could mean trouble later on.",
+                            err.__class__.__name__, err)
+                        pass
                 else:
                     gc3libs.log.info("Ignoring ssh host key file.")
 
                 self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
+
+                if self.proxy_command:
+                    proxy = paramiko.ProxyCommand(self.proxy_command)
+                    gc3libs.log.debug("Using ProxyCommand for SSH connections: %s", self.proxy_command)
+                else:
+                    proxy = None
+                    gc3libs.log.debug("Using no ProxyCommand for SSH connections.")
+
                 gc3libs.log.debug(
-                    "Connecting to host '%s' as user '%s' via SSH (timeout %ds)...",
-                    self.remote_frontend, self.username, self.timeout)
+                    "Connecting to host '%s' (port %s) as user '%s' via SSH "
+                    "(timeout %ds)...", self.remote_frontend, self.port,
+                    self.username, self.timeout)
                 self.ssh.connect(self.remote_frontend,
                                  timeout=self.timeout,
                                  username=self.username,
+                                 port=self.port,
                                  allow_agent=True,
-                                 key_filename=self.keyfile)
+                                 key_filename=self.keyfile,
+                                 sock=proxy)
                 self.sftp = self.ssh.open_sftp()
                 self._is_open = True
-        except Exception, ex:
+        except Exception as ex:
             gc3libs.log.error(
                 "Could not create ssh connection to %s: %s: %s",
-                self.remote_frontend, ex.__class__.__name__, str(ex))
+                self.remote_frontend, ex.__class__.__name__, ex)
             self._is_open = False
 
             # Try to understand why the ssh connection failed.
@@ -505,8 +669,9 @@ class SshTransport(Transport):
                     a = paramiko.Agent()
                     try:
                         running_ssh_agent = a._conn
-                    except AttributeError, ax:
-                        gc3libs.log.warning('Probably running paramiko version <= 1.7.7.2  ... ')
+                    except AttributeError:
+                        gc3libs.log.warning('Probably running paramiko '
+                                            'version <= 1.7.7.2  ... ')
                         running_ssh_agent = a.conn
 
                     if not running_ssh_agent:
@@ -533,8 +698,8 @@ class SshTransport(Transport):
                                 " command.")
 
             raise gc3libs.exceptions.TransportError(
-                "Failed while connecting to remote host '%s': %s"
-                % (self.remote_frontend, str(ex)))
+                "Failed connecting to remote host '{hostname}': {msg}"
+                .format(hostname=self.remote_frontend, msg=ex))
 
     @same_docstring_as(Transport.chmod)
     def chmod(self, path, mode):
@@ -542,7 +707,7 @@ class SshTransport(Transport):
             # check connection first
             self.connect()
             self.sftp.chmod(path, mode)
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Error changing remote path '%s' mode to 0%o: %s: %s"
                 % (path, mode, ex.__class__.__name__, str(ex)))
@@ -557,9 +722,10 @@ class SshTransport(Transport):
             gc3libs.log.debug("SshTransport running `%s`... ", command)
             stdin_stream, stdout_stream, stderr_stream = \
                 self.ssh.exec_command(command)
-            stdout = ''
-            stderr = ''
-            if not detach:
+            if detach:
+                stdout = ''
+                stderr = ''
+            else:
                 stdout = stdout_stream.read()
                 stderr = stderr_stream.read()
             exitcode = stdout_stream.channel.recv_exit_status()
@@ -567,7 +733,7 @@ class SshTransport(Transport):
                 "Executed command '%s' on host '%s'; exit code: %d"
                 % (command, self.remote_frontend, exitcode))
             return exitcode, stdout, stderr
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Failed executing remote command '%s': %s: %s"
                 % (command, ex.__class__.__name__, str(ex)))
@@ -578,18 +744,18 @@ class SshTransport(Transport):
             self.connect()
             self.sftp.stat(path)
             return True
-        except IOError, err:
+        except IOError as err:
             if err.errno == 2:
                 return False
             else:
                 raise gc3libs.exceptions.TransportError(
                     "Could not stat() file '%s' on host '%s': %s: %s"
-                    % (source, self.remote_frontend,
+                    % (path, self.remote_frontend,
                        err.__class__.__name__, str(err)))
-        except Exception, err:
+        except Exception as err:
             raise gc3libs.exceptions.TransportError(
                 "Could not stat() file '%s' on host '%s': %s: %s"
-                % (source, self.remote_frontend,
+                % (path, self.remote_frontend,
                    err.__class__.__name__, str(err)))
 
     @same_docstring_as(Transport.get_remote_username)
@@ -606,7 +772,7 @@ class SshTransport(Transport):
             self.connect()
             self.sftp.listdir(path)
             return True
-        except IOError, ex:
+        except IOError as ex:
             if ex.errno == 2:
                 return False
             else:
@@ -618,13 +784,14 @@ class SshTransport(Transport):
             # check connection first
             self.connect()
             return self.sftp.listdir(path)
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not list directory '%s' on host '%s': %s: %s"
                 % (path, self.remote_frontend, ex.__class__.__name__, str(ex)))
 
     @same_docstring_as(Transport.makedirs)
-    def makedirs(self, path, mode=0777):
+    def makedirs(self, path, mode=0o777):
+        gc3libs.log.debug("Making remote directory path '%s' ...", path)
         dirs = path.split('/')
         if '..' in dirs:
             raise gc3libs.exceptions.InvalidArgument(
@@ -649,7 +816,8 @@ class SshTransport(Transport):
         gc3libs.log.debug("SshTransport.put(): local source: '%s';"
                           " remote destination: '%s'; remote host: '%s'.",
                           source, destination, self.remote_frontend)
-        self.connect() # ensure connection is up
+        self.connect()  # ensure connection is up
+
         Transport.put(self, source, destination,
                       ignore_errors, overwrite, changed_only)
 
@@ -659,21 +827,60 @@ class SshTransport(Transport):
         """
         self.sftp.put(source, destination)
 
-    @same_docstring_as(Transport.get)
     def get(self, source, destination, ignore_nonexisting=False,
             overwrite=False, changed_only=True):
+        """
+        Copy remote file `source` to local `destination` using SFTP.
+
+        .. note::
+
+          The SSH library Paramiko_ can apparently fail downloading
+          large files, causing Python processes using it to die
+          without a traceback. For more details, see:
+          `<http://stackoverflow.com/questions/12486623/paramiko-fails-to-download-large-files-1gb>`_
+          for details.
+
+          According to user *Screwtape*'s answer, the problem lies in
+          Paramiko's ``SFTPClient.get()``, which requests all blocks
+          in the file at once; this can exhaust the system's resources
+          depending on the remote file size and local RAM.  Therefore
+          the simple code::
+
+                  self.sftp.get(source, destination)
+
+          will not work reliably in all cases, but we cannot foretell
+          *when exactly* it will break.
+
+          The approach taken here is thus the following:
+
+          - use Paramiko's ``SFTPClient.get()`` if the file size
+            exceeds a configured threshold (parameter
+            ``large_file_threshold`` given to the `SshTransport`
+            constructor);
+          - copy the file one chunk at a time, with a configurable
+            chunk size (parameter ``large_file_chunk_size`` given to
+            the `SshTransport` constructor) otherwise.
+
+          The second method is slower but more reliable so, in case of
+          doubt, it's safer to err on the "low threshold" side.
+        """
         gc3libs.log.debug("SshTranport.get(): remote source %s; "
                           "remote host: %s; local destination: %s.",
                           source, self.remote_frontend, destination)
-        self.connect() # ensure connection is up
+        # ensure connection is up
+        self.connect()
+        # delegate actual transfer to `self._get_impl()`
         Transport.get(self, source, destination,
                       ignore_nonexisting, overwrite, changed_only)
 
     def _get_impl(self, source, destination):
-        """
-        Copy remote file `source` to local `destination` using SFTP.
-        """
-        self.sftp.get(source, destination)
+        if self.stat(source).st_size < self.large_file_threshold:
+            self.sftp.get(source, destination)
+        else:
+            with self.sftp.open(source) as fsrc:
+                with open(destination, 'w') as fdst:
+                    shutil.copyfileobj(fsrc, fdst,
+                                       self.large_file_chunk_size)
 
     @same_docstring_as(Transport.remove)
     def remove(self, path):
@@ -684,7 +891,7 @@ class SshTransport(Transport):
             # check connection first
             self.connect()
             self.sftp.remove(path)
-        except IOError, ex:
+        except IOError as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not remove '%s' on host '%s': %s: %s"
                 % (path, self.remote_frontend, ex.__class__.__name__, str(ex)))
@@ -703,7 +910,7 @@ class SshTransport(Transport):
             if exit_code != 0:
                 raise Exception("Remote command '%s' failed with code %d: %s"
                                 % (_command, exit_code, stderr))
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not remove tree '%s' on host '%s': %s: %s"
                 % (path, self.remote_frontend,
@@ -714,10 +921,10 @@ class SshTransport(Transport):
         try:
             self.connect()
             return self.sftp.stat(path)
-        except Exception, err:
+        except Exception as err:
             raise gc3libs.exceptions.TransportError(
                 "Could not stat() file '%s' on host '%s': %s: %s"
-                % (source, self.remote_frontend,
+                % (path, self.remote_frontend,
                    err.__class__.__name__, str(err)))
 
     @same_docstring_as(Transport.open)
@@ -726,7 +933,7 @@ class SshTransport(Transport):
             # check connection first
             self.connect()
             return self.sftp.open(source, mode, bufsize)
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not open file '%s' on host '%s': %s: %s"
                 % (source, self.remote_frontend,
@@ -757,7 +964,6 @@ class SshTransport(Transport):
 #
 
 import subprocess
-import shutil
 
 
 class LocalTransport(Transport):
@@ -765,8 +971,29 @@ class LocalTransport(Transport):
     _is_open = False
     _process = None
 
-    def __init__(self):
-        pass
+    def __init__(self, **extra_args):
+        # logging code in class `Transport` assumes
+        # a host name is recorded into `.remote_frontend`
+        self.remote_frontend = (platform.node() or 'localhost')
+        if __debug__ and extra_args:
+            gc3libs.log.debug(
+                "LocalTransport: ignoring extra init arguments: %s",
+                ', '.join("{0}={1!r}".format(k, v)
+                          for k,v in extra_args.items()))
+
+
+    # pylint: disable=too-many-arguments,unused-argument
+    def set_connection_params(self, hostname, username=None, keyfile=None,
+                              port=None, timeout=None):
+        """
+        Set the host name stored in this `LocalTransport` instance.
+        Any other argument is ignored.
+
+        This method exists only for interface compatibility with
+        :class:`SshTranport`, which see.
+        """
+        self.remote_frontend = hostname
+
 
     def get_proc_state(self, pid):
         """
@@ -782,7 +1009,7 @@ class LocalTransport(Transport):
         TransportError if any other Exception is raised
         """
 
-        if (not self._process is None) and (_process.pid == pid):
+        if (self._process is not None) and (self._process.pid == pid):
             return self._process.poll()
         else:
 
@@ -804,43 +1031,27 @@ class LocalTransport(Transport):
 
             except IOError:
                 raise
-            except Exception, ex:
-                log.error("Error while trying to read status file. Error"
-                          " type %s. message %s" % (ex.__class__, ex.message))
-                raise gc3libs.exceptions.TransportError(x.message)
+            except Exception as ex:
+                gc3libs.log.error(
+                    "Error while trying to read status file. Error"
+                    " type %s. message %s" % (ex.__class__, ex.message))
+                raise gc3libs.exceptions.TransportError(ex.message)
 
     @same_docstring_as(Transport.connect)
     def connect(self):
-        gc3libs.log.debug("Opening LocalTransport... ")
         self._is_open = True
 
     @same_docstring_as(Transport.chmod)
     def chmod(self, path, mode):
         try:
             os.chmod(path, mode)
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Error changing local path '%s' mode to 0%o: %s: %s"
                 % (path, mode, ex.__class__.__name__, str(ex)))
 
-    def _execute_command_and_detach(self, command):
-        assert self._is_open is True, \
-            "`Transport.execute_command()` called" \
-            " on `Transport` instance closed / not yet open"
-
-        try:
-            _process = subprocess.Popen(command,
-                                        close_fds=True,
-                                        stdout=subprocess.PIPE,
-                                        shell=True)
-            return _process.pid
-        except Exception, ex:
-            raise gc3libs.exceptions.TransportError(
-                "Failed executing command '%s': %s: %s"
-                % (command, ex.__class__.__name__, str(ex)))
-
     def get_pid(self):
-        if not self._process is None:
+        if self._process is not None:
             return self._process.pid
         else:
             return -1
@@ -849,24 +1060,36 @@ class LocalTransport(Transport):
     def execute_command(self, command, detach=False):
         assert self._is_open is True, \
             "`Transport.execute_command()` called" \
-            " on `Transport` instance closed / not yet open"
+            " on closed (or not yet opened) `Transport` instance."
         if detach:
-            return self._execute_command_and_detach(command)
+            command = command + ' &'
         try:
-            self._process = subprocess.Popen(command,
-                                             stdout=subprocess.PIPE,
-                                             stderr=subprocess.PIPE,
-                                             close_fds=True, shell=True)
-            stdout, stderr = self._process.communicate()
-            exitcode = self._process.returncode
+            process = subprocess.Popen(
+                command,
+                stdout=(None if detach else subprocess.PIPE),
+                stderr=(None if detach else subprocess.PIPE),
+                close_fds=True, shell=True)
+            if detach:
+                process.wait()
+                exitcode = process.returncode
+                stdout = ''
+                stderr = ''
+            else:
+                self._process = process
+                stdout, stderr = self._process.communicate()
+                exitcode = self._process.returncode
             gc3libs.log.debug(
                 "Executed local command '%s', got exit status: %d",
                 command, exitcode)
+            # output and error streams are opened in binary mode, so
+            # we must convert them into text strings
+            stdout = to_str(stdout, 'terminal')
+            stderr = to_str(stderr, 'terminal')
             return exitcode, stdout, stderr
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Failed executing command '%s': %s: %s"
-                % (command, ex.__class__.__name__, str(ex)))
+                % (command, ex.__class__.__name__, ex))
 
     @same_docstring_as(Transport.exists)
     def exists(self, path):
@@ -917,16 +1140,17 @@ class LocalTransport(Transport):
 
         try:
             return os.listdir(path)
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not list local directory '%s': %s: %s"
                 % (path, ex.__class__.__name__, str(ex)))
 
     @same_docstring_as(Transport.makedirs)
-    def makedirs(self, path, mode=0777):
+    def makedirs(self, path, mode=0o777):
+        gc3libs.log.debug("Making directory path '%s' ...", path)
         try:
             os.makedirs(path, mode)
-        except OSError, e:
+        except OSError as e:
             if e.errno == errno.EEXIST:
                 pass
 
@@ -937,7 +1161,7 @@ class LocalTransport(Transport):
             "`Transport.put()` called" \
             " on `Transport` instance closed / not yet open"
         Transport.put(self, source, destination,
-                       ignore_errors, overwrite, changed_only)
+                      ignore_errors, overwrite, changed_only)
 
     def _put_impl(self, source, destination):
         """
@@ -954,7 +1178,7 @@ class LocalTransport(Transport):
         try:
             gc3libs.log.debug("Removing %s", path)
             return os.remove(path)
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not remove file '%s': %s: %s"
                 % (path, ex.__class__.__name__, str(ex)))
@@ -969,25 +1193,29 @@ class LocalTransport(Transport):
             gc3libs.log.debug("LocalTransport.remove_tree():"
                               " removing local directory tree '%s'" % path)
             return shutil.rmtree(path)
-        except Exception, ex:
-            raise gc3libs.exceptions.TransportError(
-                "Could not remove directory tree '%s': %s: %s"
-                % (path, ex.__class__.__name__, str(ex)))
+        except OSError as err:
+            if err.errno == errno.ENOENT:
+                # ignore "No such file or directory"
+                pass
+            else:
+                raise gc3libs.exceptions.TransportError(
+                    "Could not remove directory tree '%s': %s: %s"
+                    % (path, ex.__class__.__name__, err))
 
     @same_docstring_as(Transport.stat)
     def stat(self, path):
         try:
             return os.stat(path)
-        except Exception, err:
+        except Exception as err:
             raise gc3libs.exceptions.TransportError(
                 "Could not stat() file '%s' on host localhost: %s: %s"
-                % (source, err.__class__.__name__, str(err)))
+                % (path, err.__class__.__name__, str(err)))
 
     @same_docstring_as(Transport.open)
-    def open(self, source, mode, bufsize=0):
+    def open(self, source, mode, bufsize=-1):
         try:
             return open(source, mode, bufsize)
-        except Exception, ex:
+        except Exception as ex:
             raise gc3libs.exceptions.TransportError(
                 "Could not open file '%s' on host localhost: %s: %s"
                 % (source, ex.__class__.__name__, str(ex)))
