@@ -1,8 +1,10 @@
 #! /usr/bin/env python
-#
+
 """
+session - persistent collections of tasks
 """
-# Copyright (C) 2012, GC3, University of Zurich. All rights reserved.
+
+# Copyright (C) 2012-2019  University of Zurich. All rights reserved.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -15,24 +17,30 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+from __future__ import absolute_import, print_function, unicode_literals
+from builtins import str
 __docformat__ = 'reStructuredText'
-__version__ = '$Revision$'
 
 
 # stdlib imports
+import atexit
 import csv
+import errno
 import os
 import sys
 import shutil
+import tempfile
 
 # GC3Pie imports
 import gc3libs
 import gc3libs.exceptions
 import gc3libs.persistence
+import gc3libs.persistence.store
 import gc3libs.utils
+from gc3libs.utils import to_str
+from gc3libs.workflow import TaskCollection
 
 
 class Session(list):
@@ -59,8 +67,11 @@ class Session(list):
 
         >>> import tempfile; tmpdir = tempfile.mktemp(dir='.')
         >>> session = Session(tmpdir)
-        >>> sorted(os.listdir(tmpdir))
-        ['created', 'session_ids.txt', 'store.url']
+        >>> for name in sorted(os.listdir(tmpdir)):
+        ...   print(name)
+        created
+        session_ids.txt
+        store.url
 
     When a `Session` object is created with a `path` argument pointing
     to an existing valid session, the index of jobs is automatically
@@ -123,243 +134,161 @@ class Session(list):
     """
 
     INDEX_FILENAME = 'session_ids.txt'
-    STORE_URL_FILENAME = "store.url"
-    TIMESTAMP_FILES = {'start': 'created',
-                       'end'  : 'finished'}
+    STORE_URL_FILENAME = 'store.url'
+    TIMESTAMP_FILES = {
+        'start': 'created',
+        'end':   'finished'
+    }
 
     DEFAULT_JOBS_DIR = 'jobs'
 
-    def __init__(self, path, store_url=None, create=True, **extra_args):
+    def __init__(self, path, create=True, store_or_url=None,
+                 load=True, task_ids=None, **extra_args):
         """
         First argument `path` is the path to the session directory.
 
-        The `store_url` argument is the URL of the store, as would be
-        passed to function `gc3libs.persistence.store.make_store`:func:;
-        any additional keyword arguments are passed to `make_store` unchanged.
+        The `create` argument is used to control the behavior in case the
+        session directory does not exists already:
+
+        * If `create` is ``True`` (default) and the session directory
+          does not exists then a new session will be created.
+
+        * If `create` is ``False`` and the session directory does not
+          exists an error will be raised.
+
+        Optional argument `store_or_url` is *either* an existing valid
+        `gc3libs.persistence.store.Store` instance, *or* the URL of
+        the store, as would be passed to function
+        `gc3libs.persistence.store.make_store`:func:; any additional
+        keyword arguments are passed to `make_store` unchanged.
 
         .. note::
 
-          The optional `store_url` argument and following keyword arguments
+          The optional `store_or_url` argument and following keyword arguments
           are used if and only if a new session is being *created*; they are
           ignored when loading existing sessions!
 
-        The `create` argument is used to control the behavior in case the session does not exists.
-        If `create` is `False` and the session does not exists an error will be raised.
-        If `create` is `True` and the session does not exists then a new session will be created.
-
         By default `gc3libs.persistence.filesystem.FileSystemStore`:class:
-        (which see) is used for providing the session with a store.
+        (which see) is used for providing a new session with a store.
 
+        Finally, if optional argument `load` is ``False`` then an
+        already-existing session at `path` will be discarded and a new
+        one will be created in its place. By default, `load` is
+        ``True``, meaning that data from existing sessions is loaded
+        into memory.
         """
-        self.path = os.path.abspath(path)
+        self.path = os.path.abspath(to_str(path, 'filesystem'))
         self.name = os.path.basename(self.path)
-        self.tasks = dict()
+        self.tasks = {}
         # Session not yet created
         self.created = -1
         self.finished = -1
-        self.cmdline = extra_args.get('cmdline', None)
-
-        # check if there is an old-style session to load
-        oldstyle_index = self.path + '.csv'
-        oldstyle_store = self.path + '.jobs'
-        if (os.path.isfile(oldstyle_index) and os.path.isdir(oldstyle_store)):
-            gc3libs.log.warning(
-                "Old-style session detected in file '%s' and directory '%s'."
-                " Attempting conversion to directory based format ..."
-                % (oldstyle_index, oldstyle_store))
-            self._convert_oldstyle_session(oldstyle_index, oldstyle_store)
 
         # load or make session
-        if os.path.isdir(self.path):
+        if os.path.isdir(self.path) and load:
             # Session already exists?
             try:
-                self._load_session(**extra_args)
-            except IOError, err:
-                gc3libs.log.debug("Cannot load session '%s': %s", path, str(err))
-                if err.errno == 2:  # "No such file or directory"
-                    if create:
+                self._load_session(task_ids=task_ids, **extra_args)
+            except IOError as err:
+                gc3libs.log.debug("Cannot load session '%s': %s", path, err)
+                # "No such file or directory" => create new session
+                if err.errno == errno.ENOENT:
+                    # only create session if no task IDs were given
+                    if create and not task_ids:
                         gc3libs.log.debug(
-                            "Assuming session is incomplete or corrupted, creating it again.")
-                        self._create_session(store_url, **extra_args)
+                            "Assuming session is incomplete or corrupted,"
+                            " creating it again.")
+                        self._create_session(store_or_url, **extra_args)
                     else:
                         raise gc3libs.exceptions.InvalidArgument(
-                            "Directory '%s' does not contain a valid session" % self.path)
+                            "Directory '%s' does not contain a valid session" %
+                            self.path)
                 else:
                     raise
         else:
+            assert not task_ids, (
+                "Creating a new session with non-empty `task_ids` makes no sense!")
             if create:
-                self._create_session(store_url, **extra_args)
+                self._create_session(store_or_url, **extra_args)
             else:
                 raise gc3libs.exceptions.InvalidArgument(
-                    "Session '%s' not found" % self.path)
+                    "Session directory '%s' not found" % self.path)
 
-    def _create_session(self, store_url, **extra_args):
-        if store_url:
-            self.store_url = store_url
+    def _create_session(self, store_or_url, **extra_args):
+        if isinstance(store_or_url, gc3libs.persistence.store.Store):
+            self.store = store_or_url
         else:
-            self.store_url = os.path.join(self.path, self.DEFAULT_JOBS_DIR)
-        # Must create its directory before `make_store` is called,
-        # or SQLite raises an "OperationalError: unable to open
-        # database file None None"
-        gc3libs.utils.mkdir(self.path)
-        self.store = gc3libs.persistence.make_store(self.store_url, **extra_args)
+            if store_or_url is None:
+                store_or_url = os.path.join(self.path, self.DEFAULT_JOBS_DIR)
+            # Ensure session directory exists before `make_store` is
+            # called, or else SQLite raises an "OperationalError:
+            # unable to open database file None None"
+            gc3libs.utils.mkdir(self.path)
+            self.store = gc3libs.persistence.make_store(store_or_url, **extra_args)
+        self.store_url = self.store.url
         self._save_store_url_file()
         self._save_index_file()
 
         # Save the current command line on ``created`` file.
-        fd = open(os.path.join(
-            self.path,
-            self.TIMESTAMP_FILES['start']),'w')
-        fd.write(str.join(' ', sys.argv)+'\n')
-        fd.close()
+        with open(os.path.join(
+                self.path,
+                self.TIMESTAMP_FILES['start']), 'w') as fd:
+            fd.write(' '.join(sys.argv) + '\n')
 
         self.set_start_timestamp()
 
-    def _convert_oldstyle_session(self, index_csv, jobs_dir):
-        """
-        Convert an old-style session into a new-style one.  An
-        exception is raised if any error is encountered during the
-        conversion.
-
-        An old-style session consists of a `.csv` index file and a
-        `.jobs` directory.  Both will be moved to the location where
-        the new-style session expects them, i.e., into file
-        ``index.txt`` and subdirectory ``jobs`` respectively.
-
-        Converted sessions use a `FilesystemStore`:class: located in
-        the `.jobs` directory; any other setting of `store_url` in
-        this class instance is ignored and overwritten with the new
-        `jobs` directory location.  In other words, storage is *not*
-        converted to the any new format -- it is just relocated on the
-        filesystem.
-
-        .. fixme::
-
-          If the conversion process fails halfway through, you will
-          end up with something that is neither a valid old-style
-          session nor a valid new-style one!  This is certainly a bug.
-
-        """
-        # check access to new-style session dir and make it if needed
-        if os.path.exists(self.path):
-            gc3libs.utils.test_file(self.path, os.W_OK|os.X_OK, isdir=True)
-        else:
-            os.makedirs(self.path)
-
-        # convert job index file
-        ids_out = open(os.path.join(self.path, Session.INDEX_FILENAME), 'w')
-        ids_in = open(index_csv, 'r')
-        try:
-            for row in csv.DictReader(ids_in, ['jobname', 'persistent_id',
-                                               'state', 'info']):
-                ids_out.write(row['persistent_id'])
-                ids_out.write('\n')
-            ids_out.close()
-            ids_in.close()
-        except Exception, err:
-            gc3libs.log.error(
-                "Error converting old-style session index '%s' to new-style one '%s':"
-                " %s: %s", ids_in.name, ids_out.name, err.__class__.__name__, str(err))
-            try:
-                os.remove(ids_out.name)
-            except:
-                pass
-            ids_out.close()
-            ids_in.close()
-
-        # move jobs directory
-        new_jobs_dir = os.path.join(self.path, self.DEFAULT_JOBS_DIR)
-        try:
-            shutil.move(jobs_dir, new_jobs_dir)
-        except Exception, err:
-            gc3libs.log.error(
-                "Error moving old-style session store '%s' to new location '%s':"
-                " %s: %s", jobs_dir, new_jobs_dir, err.__class__.__name__, str(err))
-            try:
-                os.remove(ids_out.name)
-            except:
-                pass
-            # XXX: should we undo the `shutil.move` above? how?
-
-        # create `store.url` file pointing to new location
-        try:
-            self.store_url = ('file://%s' % os.path.abspath(new_jobs_dir))
-            self._save_store_url_file()
-        except Exception, err:
-            gc3libs.log.error(
-                "Error moving old-style session store '%s' to new location '%s':"
-                " %s: %s", jobs_dir, new_jobs_dir, err.__class__.__name__, str(err))
-            try:
-                os.remove(ids_out.name)
-            except:
-                pass
-            try:
-                shutil.move(new_jobs_dir, jobs_dir)
-            except Exception, err:
-                # log error and ignore it
-                gc3libs.log.error(
-                    "Additionally, got a %s while moving back '%s' to '%s': %s",
-                    err.__class__.__name__, new_jobs_dir, jobs_dir, str(err))
-        # Try to guess when the session has been started
-        statinfo = os.stat(index_csv)
-        self.created = min(statinfo.st_ctime,
-                           statinfo.st_mtime,
-                           statinfo.st_atime)
-        self.set_start_timestamp(self.created)
-
-        # if we got this far, everything is fine and we can remove the
-        # old index (and ignore any related error)
-        try:
-            os.remove(index_csv)
-        except Exception, err:
-            gc3libs.log.warning(
-                "Ignoring '%s' error that occurred in removing file '%s': %s",
-                err.__class__.__name__, index_csv, str(err))
-
-        gc3libs.log.info("Successfully converted old-style session"
-                         " to new-style session directory '%s'", self.path)
-
-    def _load_session(self, **extra_args):
+    def _load_session(self, store=None, task_ids=None, **extra_args):
         """
         Load an existing session from disk.
 
-        Keyword arguments are passed to the `make_store` factory
-        method unchanged.
+        If a `store` argument is given, it is used as the session's
+        task store; otherwise a new task store is created, passing
+        keyword arguments to the `make_store` factory method.
 
-        Any error that occurs while loading jobs from disk is ignored.
+        Any error that occurs while loading jobs from disk is ignored,
+        unless environment variable ``GC3PIE_NO_CATCH_ERRORS``
+        includes the keywords ``session`` or ``persistence``.
         """
-        try:
-            store_fname = os.path.join(self.path, self.STORE_URL_FILENAME)
-            self.store_url = gc3libs.utils.read_contents(store_fname).strip()
-        except IOError:
-            gc3libs.log.info(
-                "Unable to load session: file %s is missing." % (store_fname))
-            raise
-        self.store = gc3libs.persistence.make_store(self.store_url, **extra_args)
+        if store is not None:
+            self.store = store
+            self.store_url = self.store.url
+        else:
+            store_filename = os.path.join(self.path, self.STORE_URL_FILENAME)
+            try:
+                self.store_url = \
+                    gc3libs.utils.read_contents(store_filename).strip()
+                gc3libs.log.debug(
+                    "Loading session from URL %s ...", self.store_url)
+            except (OSError, IOError) as err:
+                gc3libs.log.error(
+                    "Unable to load session tasks from file `%s`: %s",
+                    store_filename, err)
+                raise
+            self.store = gc3libs.persistence.make_store(
+                self.store_url, **extra_args)
 
-        idx_filename = os.path.join(self.path, self.INDEX_FILENAME)
-        try:
-            idx_fd = open(idx_filename)
-            ids = idx_fd.read().split()
-            idx_fd.close()
-        except:
-            idx_fd.close()
-            raise
+        if task_ids is None:
+            idx_filename = os.path.join(self.path, self.INDEX_FILENAME)
+            try:
+                with open(idx_filename) as idx_file:
+                    task_ids = idx_file.read().split()
+            except (OSError, IOError) as err:
+                gc3libs.log.error(
+                    "Unable to load session index from file `%s`: %s",
+                    idx_filename, err)
+                raise
 
         try:
             start_file = os.path.join(
                 self.path, self.TIMESTAMP_FILES['start'])
             self.created = os.stat(start_file).st_mtime
         except OSError:
+            # FIXME: should we also set `self.created` to the current timestamp?
             gc3libs.log.warning(
-                "Unable to recover starting time from existing session: file %s is missing." % (start_file))
+                "Unable to recover starting time from existing session:"
+                " file `%s` is missing.", start_file)
 
-        for task_id in ids:
-            try:
-                self.tasks[task_id] = self.store.load(task_id)
-            except Exception, err:
-                gc3libs.log.warning(
-                    "Ignoring error from loading '%s': %s", task_id, str(err))
+        self.tasks = self.load_many(task_ids, flush=False)
 
     def destroy(self):
         """
@@ -377,7 +306,9 @@ class Session(list):
         if os.path.exists(self.path):
             shutil.rmtree(self.path)
 
-    ## collection management
+    #
+    # collection management
+    #
 
     def add(self, task, flush=True):
         """
@@ -412,43 +343,44 @@ class Session(list):
             False
 
         """
-        newid = self.store.save(task)
-        self.tasks[newid] = task
+        try:
+            task_id = task.persistent_id
+        except AttributeError:
+            task_id = self.store.save(task)
+        self.tasks[task_id] = task
         if flush:
             self.flush()
-        return newid
+        return task_id
 
     def forget(self, task_id, flush=True):
         """
         Remove task identified by `task_id` from the current session
         *but not* from the associated storage.
         """
-        if task_id not in self.tasks:
-            raise gc3libs.exceptions.InvalidArgument(
-                "Task '%s' not found in session" % task_id)
-        self.tasks.pop(task_id)
-        if flush:
+        obj = self.tasks.pop(task_id, None)
+        if flush and obj is not None:
             self.flush()
 
     def _recursive_remove_from_store(self, task_id):
         """
-        Remove a task from the store and, if the object has a `tasks` attribute containing a list of other tasks, remove them from the store
+        Remove a task from the store and, if the object has a `tasks`
+        attribute containing a list of other tasks, remove them from the store
         """
         queue = [task_id]
         while queue:
-            toremove = queue.pop()
-            obj = self.store.load(toremove)
+            id_to_remove = queue.pop()
+            obj = self.store.load(id_to_remove)
             try:
                 for child in obj.tasks:
                     queue.append(child.persistent_id)
             except AttributeError:
                 pass
             try:
-                self.store.remove(toremove)
-            except Exception, ex:
+                self.store.remove(id_to_remove)
+            except Exception as err:
                 gc3libs.log.warning(
-                    "Error removing task id `%s` from the store:"
-                    " %s" % ex)
+                    "Error removing task id `%s` from the store: %s",
+                    err)
 
     def remove(self, task_id, flush=True):
         """
@@ -462,21 +394,59 @@ class Session(list):
         return len(self.tasks)
 
     def __iter__(self):
-        return self.tasks.itervalues()
+        return iter(self.tasks.values())
+
+    class _BfsIterator(object):
+        """
+        Iterate over tasks in a group, recursively descending `TaskCollection`s.
+
+        Iteration is done in a breadth-first manner.
+        """
+        __slots__ = ('_queue')
+
+        def __init__(self, toplevel=None):
+            self._queue = []
+            if toplevel:
+                self._queue.extend(toplevel)
+
+        def __iter__(self):
+            return self
+
+        def __next__(self):
+            try:
+                task = self._queue.pop(0)
+                try:
+                    self._queue.extend(task.tasks)
+                except AttributeError:
+                    pass  # `task` is no `TaskCollection`
+                return task
+            except IndexError: # empty queue
+                raise StopIteration()
+
+        next = __next__
+
+    def iter_workflow(self):
+        """
+        Iterate over all tasks in this session.
+
+        Recursively descends `TaskCollection`:class:
+        in  a breadth-first manner.
+        """
+        return self._BfsIterator(self.tasks.values())
 
     def list_ids(self):
         """
         Return set of all task IDs belonging to this session.
         """
-        return self.tasks.keys()
+        return list(self.tasks.keys())
 
     def list_names(self):
         """
         Return set of names of tasks belonging to this session.
         """
-        return set(task.jobname for task in self.tasks.values())
+        return set(task.jobname for task in list(self.tasks.values()))
 
-    ## persistence management
+    # persistence management
 
     def flush(self):
         """
@@ -492,14 +462,49 @@ class Session(list):
         self._save_store_url_file()
         self._save_index_file()
 
-    def load(self, obj_id):
+    def load(self, obj_id, add=True, flush=True):
         """
         Load an object from persistent storage and return it.
 
         This is just a convenience proxy for calling method `load` on
         the `Store` instance associated with this session.
         """
-        return self.store.load(obj_id)
+        obj = self.store.load(obj_id)
+        if add:
+            self.add(obj, flush)
+        return obj
+
+    def load_many(self, obj_ids, add=True, flush=True):
+        """
+        Load objects given their IDs from persistent storage.
+
+        Return a dictionary mapping task ID to the actual
+        retrieved `Task`:class: object.
+        """
+        tasks = {}
+        for task_id in obj_ids:
+            try:
+                tasks[task_id] = self.load(task_id, add, flush)
+            except Exception as err:
+                if gc3libs.error_ignored(
+                        # context:
+                        # - module
+                        'session',
+                        # - class
+                        'Session',
+                        # - method
+                        'load',
+                        # - actual error class
+                        err.__class__.__name__,
+                        # - additional keywords
+                        'persistence',
+                ):
+                    gc3libs.log.warning(
+                        "Ignoring error from loading '%s': %s", task_id, err)
+                else:
+                    # propagate exception back to caller
+                    raise
+        return tasks
 
     def save(self, obj):
         """
@@ -537,7 +542,7 @@ class Session(list):
         """
         Save all modified tasks to persistent storage.
         """
-        for task in self.tasks.itervalues():
+        for task in self.tasks.values():
             if task.changed:
                 self.save(task)
         if flush:
@@ -548,15 +553,10 @@ class Session(list):
         Save job IDs to the default session index.
         """
         idx_filename = os.path.join(self.path, self.INDEX_FILENAME)
-        try:
-            idx_fd = open(idx_filename, 'w')
+        with open(idx_filename, 'w') as idx_fd:
             for task_id in self.tasks:
                 idx_fd.write(str(task_id))
                 idx_fd.write('\n')
-            idx_fd.close()
-        except:
-            idx_fd.close()
-            raise
 
     def _save_store_url_file(self):
         """
@@ -566,7 +566,7 @@ class Session(list):
         be created; if it exists, it will be overwritten.
         """
         store_url_filename = os.path.join(self.path, self.STORE_URL_FILENAME)
-        gc3libs.utils.write_contents(store_url_filename, self.store_url)
+        gc3libs.utils.write_contents(store_url_filename, str(self.store_url))
 
     def _touch_file(self, filename, time=None):
         """
@@ -592,7 +592,7 @@ class Session(list):
         creation/modification time will be used to know when the
         session has sarted.
         """
-        self.created = self._touch_file(self.TIMESTAMP_FILES['start'])
+        self.created = self._touch_file(self.TIMESTAMP_FILES['start'], time)
 
     def set_end_timestamp(self, time=None):
         """
@@ -604,10 +604,72 @@ class Session(list):
         finished, so this method should be called by a
         `SessionBasedScript`:class: class.
         """
-        self.finished = self._touch_file(self.TIMESTAMP_FILES['end'])
+        self.finished = self._touch_file(self.TIMESTAMP_FILES['end'], time)
 
 
-## main: run tests
+class TemporarySession(Session):
+    """
+    Create a session from a store URL.
+
+    In contrast with the regular `Session`:class: object, a
+    `TemporarySession`:class: does not persist any metadata about the
+    task collection.
+    In particular:
+
+    - The session index (list of task IDs belonging to the session) is
+      initialized from the entire list of jobs present in the given
+      `Store`:class: (unless a list is explicitly passed in the
+      `task_ids` argument to the constructor). This means that, unlike
+      plain `Session`:class: objects, two `TemporarySession`:class:
+      objects cannot share the same store.
+
+    - The session directory (``path`` in the `Session`:class:
+      constructor) is created on a temporary location on the
+      filesystem and deleted when the :class:`TemporarySession` is
+      destroyed.
+
+    - Timestamps will be set to the time the `TemporarySession`:class:
+      Python object is created; two `TemporarySession`:class:
+      instances with the same backing store can have different
+      creation timestamps, depending on when exactly they were
+      instanciated.
+
+    The `TemporarySession`:class: is only provided as a convenience to
+    use code that was built on top of a `Session`:class: with a
+    "naked" `Store`:class:.
+    """
+
+    def __init__(self, store_or_url, task_ids=None, delete=True, **extra_args):
+        # make temporary session dir
+        path = tempfile.mkdtemp(
+            prefix='gc3pie.TemporarySession.',
+            suffix='.d')
+        # ensure temp directory is deleted
+        if delete:
+            def cleanup():
+                if os.path.exists(path):
+                    shutil.rmtree(path)
+            atexit.register(cleanup)
+        # init `Session` class
+        extra_args.setdefault('create', True)
+        extra_args.setdefault('load', False)
+        super(TemporarySession, self).__init__(
+            path, store_or_url=store_or_url, **extra_args)
+        # populate index
+        if task_ids is None:
+            try:
+                task_ids = self.store.list()
+            except NotImplementedError:
+                raise RuntimeError(
+                    "Cannot create temporary session:"
+                    " Task store `{0}` does not support listing all task IDs.")
+        self.tasks = self.load_many(task_ids, flush=True)
+        self._save_index_file()
+        # use url as the session name
+        self.name = str(self.store.url)
+
+
+# main: run tests
 
 if "__main__" == __name__:
     import doctest
